#include "generator/GalaxyGenerator.h"
#include "utility/RandomNumberGenerator.h"
#include "math/orbitalMechanics/TheTwoBodyProblem.h"
#include "renderer/SimpleRenderer.h"
#include "renderer/CelestialRenderer.h"
#include "system/CelestialSystem.h"
#include "input/UserInput.h"
#include "input/CelestialConstants.h"
#include "testcase/CelestialTestcase.h"
#include "nbody/nbodyAlgorithm.h"
#include "utility/Utilities.h"
#include <python2.7/Python.h>

//Default Settings
int tot = 10 * 365; //10 years
float dt = 40;
int ss = 8;
int sb = 8;
bool runSim = false;
NbodyAlgorithm A = NbodyAlgorithm::AllPairsCPU;
NbodyAlgorithm B = NbodyAlgorithm::CelestialNbodyCPU;
IntegrationMethod im_A = IntegrationMethod::SymplecticEuler;
IntegrationMethod im_B = IntegrationMethod::SymplecticEuler;

void commandLine_help() {

	printf("\n");
	printf("Welcome to the Celestial Simulator!\n");
	printf("-----------------------------------------------------------------------------------------------------------------------\n");
	printf("Developed by Joakim Hommeland\n");
	printf("For more information please take a look at the accompanying thesis\n");
	printf("-----------------------------------------------------------------------------------------------------------------------\n");
	printf("Usage: \n");
	printf("CelestialSimulator -twoBody [1,2]\n");
	printf("- Runs Predefined Test case 1 and 2\n");
	printf("CelestialSimulator -compare [args]\n");
	printf("- Runs a comparison test between N-Body algorithms \n");
	printf("CelestialSimulator -performance [args]\n");
	printf("- Runs a performance test on the given algorithm\n");
	printf("CelestialSimulator [args]\n");
	printf("- Launches the specified simulation with the Celestial Renderer \n");
	printf("	Optional Arguments: \n");
	printf("	  - SolarSystems: -ss [Int] (default: 8)\n");
	printf("	  - SolarBodies: -sb [Int] (default: 8)\n");
	printf("	  - Total Simulation Time: -totalTime [Int] in days (default: 10 * 365 -> 1 year) [For use with -compare]\n");
	printf("	  - Timestep dt: -dt [float] in days (default: 40)\n");
	printf("	  - Integration Block Size: -ibs [Int] (default: 32)\n");
	printf("	  - All-Pairs GPU Block Size: -abs [Int] (default: 32)\n");
	printf("	  - Celestial N-Body GPU Block Size: -cbs [Int] (default: 32)\n");
	printf("	  - Celestial Integration Ratio: -cir [float] (default: 2)\n");
	printf("	  - Distance to Black Hole: -cd [float] (default: 100)\n");
	printf("	  - Black Hole Star Mass Ratio: -sbhr [float] (default: 10)\n");
	printf("	  - Galaxy Planes: -gp [Int] (default: 1)\n");
	printf("	  - Galaxy Planes Distance: -gpd [float] (default: 30)\n");
	printf("	  - Celestial N-Body Threshold: -threshold [float] (default: 0/OFF)\n");
	printf("	Optional Flags: \n");
	printf("	  - Enable Log: -log \n");
	printf("	  - Show Report: -showReport [For use with -twoBody] \n");
	printf("	  - Set N-Body Algorithm: set_nbody [algorithm] \n");
	printf("	     - allpairs_cpu \n");
	printf("	     - allpairs_gpu \n");
	printf("	     - celestial_cpu \n");
	printf("	     - celestial_gpu \n");
	printf("	     - celestialGrouped_gpu \n");
	printf("	  - Set Integration Method: set_integration [integrationMethod] \n");
	printf("	     - symplecticEuler \n");
	printf("	     - explicitEuler \n");
	printf("	     - celestialSymplectic \n");
	printf("	     - celestialExplicit \n");
	printf("-----------------------------------------------------------------------------------------------------------------------\n");
	printf("\n");
	printf("\n");

}

void printRuntimeInput() {

	printf("-----------------------------------------------------------------------------------\n");
	printf("Running the simulation with the following initial settings\n");
	printf("-----------------------------------------------------------------------------------\n");
	printf("Solar Systems: %d (-ss)		Solar Bodies: %d (-sb)\n", ss, sb);
	printf("Timestep %.2f (-dt)\n", dt);
	printf("N-Body Algorithm: %s (-set_nbody)\n", Utilities::NbodyEnumToText(A).c_str());
	printf("Integration Method: %s (-set_integration)\n", Utilities::IntegrationEnumToText(im_A).c_str());
	printf("Galaxy Planes: %d (-gp)\n", UserInput::galaxyPlanes);
	printf("Galaxy Planes Distance: %.2f AU (-gpd)\n", UserInput::galaxyPlanesDistance);
	printf("Threshold Switching Enabled: %d		Threshold Distance: %.2f AU (-threshold)\n", UserInput::enableThresholdSwitching,
			UserInput::distanceThreshold);
	printf("Black Hole Mass: %.2f Earth Masses (%.2f Sun Masses) (-sbhr sun_ratio)\n", UserInput::blackHoleMass,
			UserInput::blackHoleMass / CelestialConstants::sun_mass);
	printf("Black Hole Closest Distance: %.2f AU (-cd)\n", UserInput::blackHoleClosestDistance);
	printf("-----------------------------------------------------------------------------------\n");

}

void setSimulationSettings(char **argv, int i) {

	if (strcmp(argv[i], "-ibs") == 0) { //Integration Block Size
		UserInput::integrationBlockSize = atoi(argv[i + 1]);
	} else if (strcmp(argv[i], "-cbs") == 0) { //Celestial Block Size
		UserInput::celestialNbodyBlockSize = atoi(argv[i + 1]);
	} else if (strcmp(argv[i], "-abs") == 0) { //All-Pairs Block Size
		UserInput::allPairsBlockSize = atoi(argv[i + 1]);
	} else if (strcmp(argv[i], "-cir") == 0) { //Celestial Integration Ratio
		UserInput::celestialIntegrationRatio = atoi(argv[i + 1]);
	} else if (strcmp(argv[i], "-ss") == 0) { //SolarSystems
		 ss = atoi(argv[i + 1]);
	} else if (strcmp(argv[i], "-sb") == 0) { //SolarBodies
		 sb = atoi(argv[i + 1]);
	} else if (strcmp(argv[i], "-dt") == 0) {
		 dt = atof(argv[i + 1]);
	} else if (strcmp(argv[i], "-threshold") == 0) {
		 UserInput::enableThreshold(atof(argv[i+1]));
	} else if (strcmp(argv[i], "-cd") == 0) {
		 UserInput::blackHoleClosestDistance = atof(argv[i+1]);
	} else if (strcmp(argv[i], "-sbhr") == 0) {
		 UserInput::setBlackHoleMass(atof(argv[i+1]));
	} else if (strcmp(argv[i], "-gp") == 0) {

		int num = atoi(argv[i+1]);
		if (num >= 2) {
			 UserInput::galaxyPlanes = 2;
		} else {
			 UserInput::galaxyPlanes = 1;
		}

	} else if (strcmp(argv[i], "-gpd") == 0) {
		 UserInput::galaxyPlanesDistance = atof(argv[i+1]);
	}

}

NbodyAlgorithm resolveNbodyAlgorithm(char *argv) {

	NbodyAlgorithm alg = NbodyAlgorithm::AllPairsCPU;

	if (strcmp(argv, "-allpairs_cpu") == 0) {
		alg = NbodyAlgorithm::AllPairsCPU;
	} else if (strcmp(argv, "allpairs_gpu") == 0) {
		alg = NbodyAlgorithm::AllPairsGPU;
	} else if (strcmp(argv, "celestial_cpu") == 0) {
		alg = NbodyAlgorithm::CelestialNbodyCPU;
	} else if (strcmp(argv, "celestial_gpu") == 0) {
		alg = NbodyAlgorithm::CelestialNbodyGPU;
	} else if (strcmp(argv, "celestialGrouped_gpu") == 0) {
		alg = NbodyAlgorithm::CelestialNbodyGroupedGPU;
	}

	return alg;

}

IntegrationMethod resolveIntegrationMethod(char *argv) {

	IntegrationMethod met = IntegrationMethod::SymplecticEuler;

	if (strcmp(argv, "symplecticEuler") == 0) {
		met = IntegrationMethod::SymplecticEuler;
	} else if (strcmp(argv, "explicitEuler") == 0) {
		met = IntegrationMethod::ExplicitEuler;
	} else if (strcmp(argv, "celestialSymplectic") == 0) {
		met = IntegrationMethod::CelestialIntegrationSymplectic;
	} else if (strcmp(argv, "celestialExplicit") == 0) {
		met = IntegrationMethod::CelestialIntegrationExplicit;
	}

	return met;

}

int main(int argc, char **argv) {

	if (argc == 2 && (strcmp(argv[1], "-help") == 0 || strcmp(argv[1], "-h") == 0)) { //Help
		commandLine_help();
	} else if (argc > 1 && strcmp(argv[1], "-twoBody") == 0) { //Two Body analytical and simulation comparison test

		bool showReport = false;
		float dt = 1.0;

		for (int i=3;i<argc;i++) {
		    if (i + 1 != argc) { //One argument functions

				 if (strcmp(argv[i], "-dt") == 0) { //SolarSystems
					 dt = atof(argv[i + 1]);
				 }

			}

		    if (strcmp(argv[i], "-showReport") == 0) {
				showReport = true;
			}

		}

		if (argc > 2 && atoi(argv[2]) == 1) {
			printf("Two Body Test Case\n");
			TheTwoBodyProblem::runTwoBodyTestCase(showReport, dt);
		} else if (argc > 2 && atoi(argv[2]) == 2) {
			printf("Two Body Solar System Test Case\n");
			TheTwoBodyProblem::runSolarSystemTestCase(showReport, dt);
		} else {
			commandLine_help();
		}

	} else if (argc > 1 && strcmp(argv[1], "-compare") == 0){ //Comparison of Algorithms

		bool showStepInfo = false;
		bool verbose = false;

		for (int i=1;i<argc; i++) {

			if (strcmp(argv[i], "-runSim") == 0) {
				runSim = true;
			} else if (strcmp(argv[i], "-log") == 0) {
				UserInput::logSimulation = true;
			} else if (strcmp(argv[i], "-info") == 0) {
				showStepInfo = true;
			} else if (strcmp(argv[i], "-verbose") == 0) {
				verbose = true;
			}
			if (i + 1 != argc) { //One argument functions

				 if (strcmp(argv[i], "-totalTime") == 0) {
					 tot = atoi(argv[i + 1]);
				 } else {
					 setSimulationSettings(argv, i);
				 }
			}

			if (i + 2 < argc) {
				if (strcmp(argv[i], "-set_nbody") == 0) {
					A = resolveNbodyAlgorithm(argv[i + 1]);
					B = resolveNbodyAlgorithm(argv[i + 2]);
				} else if (strcmp(argv[i], "-set_integration") == 0) {
					im_A = resolveIntegrationMethod(argv[i + 1]);
					im_B = resolveIntegrationMethod(argv[i + 2]);
				}
			}


		}

		int steps = tot / dt;
		bool showGraph = true;

		CelestialTestcase::nbodyAccelerationComparisonTest(A, B, im_A, im_B, ss, sb, dt, steps, verbose, showStepInfo, showGraph);

		if (runSim) {

			printRuntimeInput();

			RandomNumberGenerator::setSrand(0);
			printf("Running First simulation: %s\n", Utilities::NbodyEnumToText(A).c_str());
			bool useGPU_A = UserInput::setNbodyAlgorithm(A);
			CelestialSystem *system = new CelestialSystem(ss, sb, useGPU_A);
			CelestialRenderer *renderer = new CelestialRenderer(system, dt);
			renderer->setupAndRun(UserInput::res_width, UserInput::res_height);

			printRuntimeInput();

			RandomNumberGenerator::setSrand(0);
			printf("Running Second simulation: %s\n", Utilities::NbodyEnumToText(B).c_str());
			bool useGPU_B = UserInput::setNbodyAlgorithm(B);
			CelestialSystem *system2 = new CelestialSystem(ss, sb, useGPU_B);
			CelestialRenderer *renderer2 = new CelestialRenderer(system2, dt);
			renderer2->setupAndRun(UserInput::res_width, UserInput::res_height);
		}

	} else if (argc > 1 && strcmp(argv[1], "-performance") == 0) {

		for (int i=1;i<argc; i++) {

			if (strcmp(argv[i], "-log") == 0) {
				UserInput::logSimulation = true;
			}

			if (i + 1 != argc) {
				 setSimulationSettings(argv, i);

				 if (strcmp(argv[i], "-set_nbody") == 0) {
					 A = resolveNbodyAlgorithm(argv[i + 1]);
				 } else if (strcmp(argv[i], "-set_integration") == 0) {
					 im_A = resolveIntegrationMethod(argv[i + 1]);
				 }
			}

		}

		CelestialTestcase::nbodyPerformanceTest(A, im_A, ss, sb, dt);

	} else {

		for (int i=1;i<argc; i++) {

			if (strcmp(argv[i], "-log") == 0) {
				UserInput::logSimulation = true;
			}

			if (i + 1 != argc) {
				 setSimulationSettings(argv, i);

				 if (strcmp(argv[i], "-set_nbody") == 0) {
					 A = resolveNbodyAlgorithm(argv[i + 1]);
				 } else if (strcmp(argv[i], "-set_integration") == 0) {
					 im_A = resolveIntegrationMethod(argv[i + 1]);
				 }
			}

		}

		printRuntimeInput();

		bool useGPU_A = UserInput::setupSimulation(A, im_A);
		CelestialSystem *system = new CelestialSystem(ss, sb, useGPU_A);
		CelestialRenderer *renderer = new CelestialRenderer(system, dt);
		renderer->setupAndRun(UserInput::res_width, UserInput::res_height);

	}

	cudaDeviceReset();

	return EXIT_SUCCESS;
}
