#include "GalaxyGenerator.h"
#include "math.h"
#include "../input/CelestialConstants.h"
#include <stdio.h>
#include "../utility/RandomNumberGenerator.h"
#include "../utility/CudaErrorCheck.h"
#include "../input/UserInput.h"

#define ToRadian(x) ((x) * M_PI / 180.0f)
#define ToDegree(x) ((x) * 180.0f / M_PI)

void GalaxyGenerator::createGalaxy(int solarSystems, int solarBodies) { //Currently only supports 2 planes (the angle is wrong)

	float blackHoleMass = UserInput::blackHoleMass;
	float starMass = CelestialConstants::sun_mass;
	float startAngle = ToRadian(45);

	//Set Galaxy Center properties
	velocities[0] = make_float3(0, 0, 0);
	positions[0] = make_float3(0, 0, 0);
	mass[0] = blackHoleMass;

	float distance = UserInput::blackHoleClosestDistance;
	float planes_distance = UserInput::galaxyPlanesDistance;

	int planes = UserInput::galaxyPlanes;

	int solarPerPlane = solarSystems / planes;
	float inclination_d =  2*M_PI / solarPerPlane;
	float azimuth_d = M_PI / planes;

	int starIndex = 1;
	for (int plane_i = 0; plane_i < planes; plane_i++) {

		float azimuth = plane_i * azimuth_d + startAngle;
		for (int solar_i = 0; solar_i < solarPerPlane; solar_i++) {

			float inclination = solar_i * inclination_d;

			//Set mass of the sun
			mass[starIndex] = starMass;

			//Step 1: Find the Star's start position in relation to the Black Hole
			positions[starIndex] = findSolarStartPosition(distance, inclination, azimuth);

			//Step 2: Find the Star's star velocity needed in order to orbit the Black Hole
			findStartVelocity(0, starIndex, 0, distance, inclination, azimuth);

			//Step 3: Find Solar System bodies start position and velocity in relation to star
			setSolarSystemValues(starIndex, starIndex + solarBodies, azimuth);

			starIndex += solarBodies;

		}

		distance += planes_distance;

	}

}

void GalaxyGenerator::setSolarSystemValues(int startIndex, int endIndex, float azimuth) {

	float planetMass = CelestialConstants::earth_mass;
	float e = CelestialConstants::earth_e;
	float a = CelestialConstants::earth_a;

	for (int i = startIndex + 1; i < endIndex; i++) {

		//Set mass of the planet
		mass[i] = planetMass;

		float theta = RandomNumberGenerator::getRand(0, 2*M_PI);

		//Find Start Position
		findPlanetStartPosition(startIndex, i, e, a, theta, azimuth);

		//Find Start Velocity
		findStartVelocity(startIndex, i, e, a, theta, azimuth);

		a += CelestialConstants::earth_a;

	}

}

float3 GalaxyGenerator::findSolarStartPosition(float r, float inclination, float azimuth) {

	return make_float3(r * sin(inclination) * cos(azimuth), r * sin(inclination) * sin(azimuth), r * cos(inclination));

}

void GalaxyGenerator::findPlanetStartPosition(int primaryIndex, int secondaryIndex, float e, float a, float theta, float azimuth) {

	float3 primaryPosition = positions[primaryIndex];
	float3 blackholePosition = positions[0];

	float3 ss_vec = make_float3(blackholePosition.x - primaryPosition.x, blackholePosition.y - primaryPosition.y,
			blackholePosition.z - primaryPosition.z);

	//Find position
	float r = a * (1 - e * e) / (1 + e * cos(theta));
	float3 positionZY = make_float3(0, r * sin(theta), r * cos(theta));

	//Rotate to match the current plane
	float3 position;
	position.x = positionZY.x * cos(2 * M_PI - azimuth) - positionZY.y * sin(2*M_PI - azimuth);
	position.y = positionZY.x * sin(2 * M_PI - azimuth) + positionZY.y * cos(2*M_PI - azimuth);
	position.z = positionZY.z;

	position.x -= ss_vec.x;
	position.y -= ss_vec.y;
	position.z -= ss_vec.z;

	positions[secondaryIndex] = position;

}

void GalaxyGenerator::findStartVelocity2D(int primaryIndex, int secondaryIndex, float e, float a, float theta) {

	float primaryMass = mass[primaryIndex];
	float secondaryMass = mass[secondaryIndex];

	//Find phi
	float phi = G * (primaryMass + secondaryMass);

	//Find Orbital Period
	float T = sqrt((4 * M_PI * M_PI / phi) * a * a * a);

	//Find Mean Motion
	float n = 2 * M_PI / T;

	//Find start velocity
	float startX = (-1) * (n * a / sqrt(1 - e * e)) * sin(theta);
	float startY = (n * a / sqrt(1 - e * e)) * (e + cos(theta));



	float3 startVelocity = make_float3(startX, startY, 0);

	velocities[secondaryIndex] = make_float3(startVelocity.x + velocities[primaryIndex].x,
			startVelocity.y + velocities[primaryIndex].y,
			startVelocity.z + velocities[primaryIndex].z);

}

void GalaxyGenerator::findStartPosition2D(int primaryIndex, int secondaryIndex, float e, float a, float theta) {

	//Find position
	float r = a * (1 - e * e) / (1 + e * cos(theta));
	float3 startPos = make_float3(r * cos(theta), r * sin(theta), 0);

	positions[secondaryIndex] = make_float3(startPos.x + positions[primaryIndex].x,
			startPos.y + positions[primaryIndex].y,
			startPos.z + positions[primaryIndex].z);

}

void GalaxyGenerator::findStartVelocity(int primaryIndex, int secondaryIndex,
		float e, float a, float theta, float azimuth) {

	float primaryMass = mass[primaryIndex];
	float secondaryMass = mass[secondaryIndex];

	//Find phi
	float phi = G * (primaryMass + secondaryMass);

	//Find Orbital Period
	float T = sqrt((4 * M_PI * M_PI / phi) * a * a * a);

	//Find Mean Motion
	float n = 2 * M_PI / T;

	//Find start velocity
	float startX = (-1) * (n * a / sqrt(1 - e * e)) * sin(theta);
	float startY = (n * a / sqrt(1 - e * e)) * (e + cos(theta));
	float3 velocityZY = make_float3(0, startY, startX);

	//Rotate it to match the plane
	float3 velocity;
	velocity.x = velocityZY.x * cos(2 * M_PI - azimuth) - velocityZY.y * sin(2 * M_PI - azimuth);
	velocity.y = velocityZY.x * sin(2 * M_PI - azimuth) + velocityZY.y * cos(2 * M_PI - azimuth);
	velocity.z = velocityZY.z;

	velocity.x += velocities[primaryIndex].x;
	velocity.y += velocities[primaryIndex].y;
	velocity.z += velocities[primaryIndex].z;

	velocities[secondaryIndex] = velocity;

}

GalaxyGenerator::GalaxyGenerator(int solarSystems, int solarBodies) {

	this->solarSystems = solarSystems;
	this->solarBodies = solarBodies;
	this->N = 1 + solarSystems * solarBodies;

	allocateVariables(N);

	createGalaxy(solarSystems, solarBodies);


}

void GalaxyGenerator::allocateVariables(int N) {

	positions = (float3*) calloc(N, sizeof(float3));
	accelerations = (float3*) calloc(N, sizeof(float3));
	velocities = (float3*) calloc(N, sizeof(float3));
	mass = (float*) calloc(N, sizeof(float));

	//Set acceleration to zero
	for (int i=0;i<N;i++) {
		accelerations[i] = make_float3(0,0,0);
	}

}

float3 *GalaxyGenerator::getPositionsGPU() {

	float3 *positions_d;

	CudaSafeCall(cudaMalloc(&positions_d, sizeof(float3) * N));

	CudaSafeCall(
			cudaMemcpy(positions_d, positions, sizeof(float3) * N,
					cudaMemcpyHostToDevice));

	free(positions);

	return positions_d;

}

float3 *GalaxyGenerator::getVelocitiesGPU() {

	float3 *velocities_d;

	CudaSafeCall(cudaMalloc(&velocities_d, sizeof(float3) * N));

	CudaSafeCall(
			cudaMemcpy(velocities_d, velocities, sizeof(float3) * N,
					cudaMemcpyHostToDevice));


	free(velocities);

	return velocities_d;

}

float3 *GalaxyGenerator::getAccelerationsGPU() {

	float3* accelerations_d;

	CudaSafeCall(cudaMalloc(&accelerations_d, sizeof(float3) * N));

	CudaSafeCall(
			cudaMemcpy(accelerations_d, accelerations,
					sizeof(float3) * N, cudaMemcpyHostToDevice));

	free(accelerations);

	return accelerations_d;

}

float *GalaxyGenerator::getMassesGPU() {

	float *masses_d;

	CudaSafeCall(cudaMalloc(&masses_d, sizeof(float) * N));

	CudaSafeCall(
			cudaMemcpy(masses_d, mass, sizeof(float) * N,
					cudaMemcpyHostToDevice));

	free(mass);

	return masses_d;

}

float3* GalaxyGenerator::getPositions() {

	return positions;

}

float3* GalaxyGenerator::getVelocities() {

	return velocities;

}

float3 *GalaxyGenerator::getAccelerations() {

	return accelerations;

}
float *GalaxyGenerator::getMasses() {

	return mass;

}

int GalaxyGenerator::getN() {

	return N;

}

int GalaxyGenerator::getSolarSystems() {

	return solarSystems;
}

int GalaxyGenerator::getSolarBodies() {

	return solarBodies;

}



