#include "SystemGenerator.h"
#include "../input/CelestialConstants.h"
#include "../nbody/NbodyInteractions.h"

/* Getters and Setters */
float3* SystemGenerator::getPositions() {
	return positions;
}

float3* SystemGenerator::getAccelerations() {
	return accelerations;
}

float3* SystemGenerator::getVelocities() {
	return velocities;
}

float* SystemGenerator::getMasses() {
	return masses;
}

int SystemGenerator::getN() {
	return N;
}

/* Constructors */

SystemGenerator::SystemGenerator(TheTwoBodyProblem *twoBody) {

	this->N = 2;

	allocateVariables(N);

	positions[1].x = twoBody->getPlanetStartPosition().x;
	positions[1].y = twoBody->getPlanetStartPosition().y;
	positions[1].z = twoBody->getPlanetStartPosition().z;

	velocities[0].x = 0;
	velocities[0].y = 0;
	velocities[0].z = 0;

	velocities[1].x = twoBody->getPlanetStartVelocity().x;
	velocities[1].y = twoBody->getPlanetStartVelocity().y;
	velocities[1].z = twoBody->getPlanetStartVelocity().z;

	//accelerations[1] = computeBodyBody(positions[1], masses[1], positions[0], masses[0], accelerations[1]);

	masses[1] = twoBody->getPlanetMass();
	masses[0] = twoBody->getStarMass();

}

SystemGenerator::SystemGenerator(std::vector<TheTwoBodyProblem*> bodies) {

	this->N = 1 + bodies.size();

	allocateVariables(N);

	//Set the position velocity and mass of the sun
	positions[0] = make_float3(0, 0, 0);
	velocities[0] = make_float3(0,0,0);
	masses[0] = bodies.front()->getStarMass();
	accelerations[0] = make_float3(0,0,0);

	double3 b, b_v;
	for (int i=1; i<N; i++) {

		TheTwoBodyProblem *body = bodies[i-1];

		b = body->getPlanetStartPosition();
		positions[i] = make_float3(b.x, b.y, b.z);

		b_v = body->getPlanetStartVelocity();
		velocities[i] = make_float3(b_v.x, b_v.y, b_v.z);

		masses[i] = body->getPlanetMass();

		accelerations[i] = make_float3(0,0,0);
	}

}

void SystemGenerator::allocateVariables(int N) {

	positions = (float3*) calloc(N, sizeof(float3));
	accelerations = (float3*) calloc(N, sizeof(float3));
	velocities = (float3*) calloc(N, sizeof(float3));
	masses = (float*) calloc(N, sizeof(float));

}

void SystemGenerator::setPlanetSolarOrbit(int planetIndex, float planetMass, float a, float e, float startTheta) {

	TheTwoBodyProblem *twoBody = TheTwoBodyProblem::initializeTwoBodySystem(masses[0], planetMass, e, a, startTheta, -1);
	positions[planetIndex].x = twoBody->getPlanetStartPosition().x;
	positions[planetIndex].y = twoBody->getPlanetStartPosition().y;
	positions[planetIndex].z = twoBody->getPlanetStartPosition().z;

	velocities[planetIndex].x = twoBody->getPlanetStartVelocity().x;
	velocities[planetIndex].y = twoBody->getPlanetStartVelocity().y;
	velocities[planetIndex].z = twoBody->getPlanetStartVelocity().z;

	accelerations[planetIndex] = computeBodyBody(positions[1], masses[1], positions[0], masses[0], accelerations[1]);

	masses[planetIndex] = twoBody->getPlanetMass();
}

SystemGenerator* SystemGenerator::generateSolarSystem(TheTwoBodyProblem *twoBody) {

	SystemGenerator *system = new SystemGenerator(twoBody);

	return system;
}

SystemGenerator* SystemGenerator::generateSolarSystem(std::vector<TheTwoBodyProblem*> bodies) {

	SystemGenerator *system = new SystemGenerator(bodies);

	return system;

}

