#include "SymplecticEuler.h"
#include "../../utility/CudaErrorCheck.h"
#include "../../input/UserInput.h"
#include "../../nbody/NbodyAllPairs.h"
#include "../../nbody/CelestialNbody.h"
#include "../../nbody/CelestialNbodyGrouped.h"
#include "../../nbody/nbodyAlgorithm.h"
#include "../../nbody/NbodyInteractions.h"
#include "../../system/CelestialSystem.h"

bool SymplecticEuler::performIntegrationCPU(CelestialSystem* system, float dt) {

	float3 *acceleration = system->getAccelerations();
	float3 *position = system->getPositions();
	float *mass = system->getMasses();
	int N = system->getN();
	int solarSystems = system->getSolarSystems();
	int solarBodies = system->getSolarBodies();

	//Update Acceleration
	float thresholdReached = system->thresholdIsReached();
	if (system->getCurrentlyEnabledAlgorithm() == NbodyAlgorithm::CelestialNbodyCPU) {
		thresholdReached = CelestialNbody::celestialNbodyCPU(acceleration, position, mass, solarSystems, solarBodies, system->thresholdSwitchingIsEnabled());
	} else { //Default to allpairs
		nbody_allpairs_cpu(acceleration, position, mass, N);
	}

	SymplecticEuler::performSymplecticIntegrationCPU(system, 0, dt);

	return thresholdReached;

}

bool SymplecticEuler::performCelestialIntegrationCPU(CelestialSystem* system, float dt) {

	if (UserInput::logSimulation) {
		printf("[Log] Celestial Integration CPU will be performed! Ratio: %f\n", UserInput::celestialIntegrationRatio);
	}

	float3 *velocity = system->getVelocities();
	float3 *acceleration = system->getAccelerations();
	float3 *position = system->getPositions();
	float *mass = system->getMasses();
	int N = system->getN();
	int solarSystems = system->getSolarSystems();
	int solarBodies = system->getSolarBodies();

	float internal_steps = UserInput::celestialIntegrationRatio;
	float external_dt = dt;
	float internal_dt = external_dt / internal_steps;

	float4 *massCenters = calculateMassCenterForSolarSystemsCPU(position, mass, solarSystems, solarBodies);
	float3 *acceleration_ext = (float3*) calloc(N, sizeof(float3));
	bool thresholdReached = CelestialNbody::calculateExternalAccelerationCPU(acceleration_ext, position, mass, massCenters,
			solarSystems, solarBodies, false, UserInput::distanceThreshold);
	free(massCenters);

	for (int i=0;i<(int)internal_steps;i++) {

		CelestialNbody::calculateInternalAccelerationCPU(acceleration, position, mass, solarSystems, solarBodies);
		SymplecticEuler::performSymplecticIntegrationCPU(system, acceleration_ext, internal_dt);

	}

	free(acceleration_ext);

	return thresholdReached;

}


bool SymplecticEuler::performCelestialIntegrationGPU(CelestialSystem* system, float dt) {

	if (UserInput::logSimulation) {
		printf("[Log] Celestial Integration GPU will be performed! Ratio: %f\n", UserInput::celestialIntegrationRatio);
	}

	float3 *velocity = system->getVelocities();
	float3 *acceleration = system->getAccelerations();
	float3 *position = system->getPositions();
	float *mass = system->getMasses();
	int N = system->getN();
	int solarSystems = system->getSolarSystems();
	int solarBodies = system->getSolarBodies();

	int ibs = UserInput::integrationBlockSize;
	int inb = N / ibs + ((N % ibs == 0) ? 0 : 1);

	int cbs = solarBodies;
	int cnb = solarSystems;
	int sharedMemSize = ibs * sizeof(float3) + ibs * sizeof(float);

	float internal_steps = UserInput::celestialIntegrationRatio;
	float external_dt = dt;
	float internal_dt = external_dt / internal_steps;

	float4* massCenters;
	CudaSafeCall( cudaMalloc(&massCenters, sizeof(float) * 4 * solarSystems) );
	calculateMassCenterForSolarSystems<<<solarSystems, solarBodies>>>(position, mass, massCenters);
	CudaCheckError();

	bool *thresholdReachedGPU;
	CudaSafeCall( cudaMalloc(&thresholdReachedGPU, sizeof(bool)) );
	CudaCheckError();

	//Save acceleration for external systems
	float3* acceleration_ext;
	CudaSafeCall( cudaMalloc(&acceleration_ext, sizeof(float) * 3 * N) );
	calculateExternalAccelerationGPU<<<inb, ibs>>>(acceleration_ext, position, mass, massCenters,
			solarSystems, solarBodies, false, UserInput::distanceThreshold, thresholdReachedGPU);

	//Perform Internal calculation internal_steps-1 times
	for (int i=0;i<(int)internal_steps;i++) {

		calculateInternalAccelerationGPU<<<cnb, cbs, sharedMemSize>>>(acceleration, position, mass,  0, N);
		CudaCheckError();

		symplecticIntegrationKernel<<<inb, ibs>>>(acceleration, acceleration_ext, velocity, position, N, internal_dt);
		CudaCheckError();

	}

	bool thresholdReached_output = system->thresholdIsReached();
	if (system->thresholdSwitchingIsEnabled()) {
		bool *thresholdReached = (bool*) malloc(sizeof(bool));
		CudaSafeCall(
				cudaMemcpy(thresholdReached, thresholdReachedGPU, sizeof(bool),
						cudaMemcpyDeviceToHost));
		thresholdReached_output = *thresholdReached;
	}

	CudaSafeCall( cudaFree(massCenters) );
	CudaSafeCall( cudaFree(thresholdReachedGPU) );
	CudaSafeCall( cudaFree(acceleration_ext) );

	return thresholdReached_output;

}

bool SymplecticEuler::performIntegrationGPU(CelestialSystem* system, float dt) {

	if (UserInput::logSimulation) {
		printf("[Log] Symplectic Euler GPU Integration will be performed!\n");
	}

	float3 *velocity = system->getVelocities();
	float3 *acceleration = system->getAccelerations();
	float3 *position = system->getPositions();
	float *mass = system->getMasses();
	int N = system->getN();
	int solarSystems = system->getSolarSystems();
	int solarBodies = system->getSolarBodies();
	bool enableThresholdSwitching = system->thresholdSwitchingIsEnabled();

	//Update Accelerations
	bool thresholdReached = system->thresholdIsReached();
	if (system->getCurrentlyEnabledAlgorithm() == NbodyAlgorithm::CelestialNbodyGPU) {
		thresholdReached = CelestialNbody::launchCelestialNbodyKernel(acceleration, position, mass, solarSystems, solarBodies, enableThresholdSwitching);
	} else if (system->getCurrentlyEnabledAlgorithm() == NbodyAlgorithm::CelestialNbodyGroupedGPU) {
		thresholdReached = CelestialNbodyGrouped::launchCelestialNbodyKernel(acceleration, position, mass, solarSystems, solarBodies, enableThresholdSwitching);
	} else { //Default to allpairs
		nbody_allpairs_gpu(acceleration, position, mass, N);
	}

	//Launch Configuration
	int block_size = UserInput::integrationBlockSize;
	int num_blocks = (N-1) / block_size + (((N-1) % block_size == 0) ? 0 : 1);

	//Perform Integration
	symplecticIntegrationKernel<<<num_blocks, block_size>>>(acceleration, 0, velocity, position, N, dt);
	CudaCheckError();

	return thresholdReached;

}

void SymplecticEuler::performSymplecticIntegrationCPU(CelestialSystem* system, float3 *acceleration_ext, float dt) {

	if (UserInput::logSimulation) {
		printf("[Log] Symplectic Euler CPU Integration will be performed!\n");
	}

	float3 *velocity = system->getVelocities();
	float3 *acceleration = system->getAccelerations();
	float3 *position = system->getPositions();
	int N = system->getN();

	float3 position_next, velocity_next;
	for (int i=1;i<N;i++) {

		//Update velocity
		velocity_next.x = velocity[i].x + acceleration[i].x * dt;
		velocity_next.y = velocity[i].y + acceleration[i].y * dt;
		velocity_next.z = velocity[i].z + acceleration[i].z * dt;


		if (acceleration_ext != 0) {
			velocity_next.x += acceleration_ext[i].x * dt;
			velocity_next.y += acceleration_ext[i].y * dt;
			velocity_next.z +=  acceleration_ext[i].z * dt;
		}


		//Update position
		position_next.x = position[i].x + velocity_next.x * dt;
		position_next.y = position[i].y + velocity_next.y * dt;
		position_next.z = position[i].z + velocity_next.z * dt;

		//Write the values back to the arrays
		velocity[i] = velocity_next;
		position[i] = position_next;

	}

}

__global__ void symplecticIntegrationKernel(float3* acceleration, float3 *acceleration_ext, float3* velocity, float3 *position, int N, float dt) {

	int threadID = blockIdx.x * blockDim.x + threadIdx.x + 1;

	if (threadID < N) {

		float3 position_next, velocity_next;

		//Update velocity
		velocity_next.x = velocity[threadID].x + acceleration[threadID].x * dt;
		velocity_next.y = velocity[threadID].y + acceleration[threadID].y * dt;
		velocity_next.z = velocity[threadID].z + acceleration[threadID].z * dt;

		//Add acceleration_ext if needed
		if (acceleration_ext != 0) {
			velocity_next.x += acceleration_ext[threadID].x * dt;
			velocity_next.y += acceleration_ext[threadID].y * dt;
			velocity_next.z +=  acceleration_ext[threadID].z * dt;
		}

		//Update position
		position_next.x = position[threadID].x + velocity_next.x * dt;
		position_next.y = position[threadID].y + velocity_next.y * dt;
		position_next.z = position[threadID].z + velocity_next.z * dt;

		//Write the values back to the arrays

		velocity[threadID] = velocity_next;
		position[threadID] = position_next;

	}

}

