/*
 * TheTwoBodyProblem.cpp
 *
 *  Created on: May 18, 2015
 *      Author: Joakim
 */

#include "TheTwoBodyProblem.h"
#include <math.h>
#include <stdio.h>
#include "../../input/UserInput.h"
#include "../../input/CelestialConstants.h"
#include "../../renderer/SimpleRenderer.h"
#include "../../nbody/nbodyAlgorithm.h"

/* Test Functions */

void TheTwoBodyProblem::runTwoBodyTestCase(bool shouldShowSimulationReport, float dt) {

	double mass_sun = CelestialConstants::sun_mass;
	double mass_planet = CelestialConstants::earth_mass;

	UserInput::nbodyAlgorithm = NbodyAlgorithm::AllPairsCPU;
	UserInput::integrationMethod = IntegrationMethod::SymplecticEuler;

	//Input Format: starMass, planetMass, Eccentricity, a, startTheta
	runTwoBodySimulationWithRenderer(mass_sun, mass_planet,
			CelestialConstants::earth_e, CelestialConstants::earth_a, 0, dt, shouldShowSimulationReport);

}

void TheTwoBodyProblem::runSolarSystemTestCase(bool shouldShowSimulationReport, float dt) {

	std::vector<TheTwoBodyProblem*> bodies = createSolarSystem(dt);

	UserInput::nbodyAlgorithm = NbodyAlgorithm::AllPairsCPU;
	UserInput::integrationMethod = IntegrationMethod::SymplecticEuler;

	SimpleRenderer *renderer = new SimpleRenderer(bodies, true, true, dt, shouldShowSimulationReport);
	renderer->setupAndRun(800, 800);

}

std::vector<TheTwoBodyProblem*> TheTwoBodyProblem::createSolarSystem(float dt) {

	std::vector<TheTwoBodyProblem*> bodies;

		double mercuryStartPos = 0;
		TheTwoBodyProblem *sunMercurySystem =
				TheTwoBodyProblem::initializeTwoBodySystem(
						CelestialConstants::sun_mass,
						CelestialConstants::mercury_mass, CelestialConstants::mercury_e,
						CelestialConstants::mercury_a, mercuryStartPos, dt);
		bodies.push_back(sunMercurySystem);

		double venusStartPos = 0;
		TheTwoBodyProblem *sunVenusSystem =
				TheTwoBodyProblem::initializeTwoBodySystem(
						CelestialConstants::sun_mass,
						CelestialConstants::venus_mass, CelestialConstants::venus_e,
						CelestialConstants::venus_a, venusStartPos, dt);
		bodies.push_back(sunVenusSystem);

		double earthStartPos = 0;
		TheTwoBodyProblem *sunEarthSystem =
				TheTwoBodyProblem::initializeTwoBodySystem(
						CelestialConstants::sun_mass,
						CelestialConstants::earth_mass, CelestialConstants::earth_e,
						CelestialConstants::earth_a, earthStartPos, dt);
		bodies.push_back(sunEarthSystem);

		double marsStartPos = 0;
		TheTwoBodyProblem *sunMarsSystem =
				TheTwoBodyProblem::initializeTwoBodySystem(
						CelestialConstants::sun_mass,
						CelestialConstants::mars_mass, CelestialConstants::mars_e,
						CelestialConstants::mars_a, marsStartPos, dt);
		bodies.push_back(sunMarsSystem);

		return bodies;
}

void TheTwoBodyProblem::runTwoBodySimulationWithRenderer(double starMass,
		double planetMass, double e, double a, double startTheta, double dt, bool shouldShowSimulationReport) {

	TheTwoBodyProblem *twoBody = TheTwoBodyProblem::initializeTwoBodySystem(
			starMass, planetMass, e, a, startTheta, dt);
	twoBody->setTimestep(dt);

	SimpleRenderer *renderer = new SimpleRenderer(twoBody, dt, shouldShowSimulationReport);
	renderer->setupAndRun(800, 800);

	delete renderer;
	delete twoBody;

}

void TheTwoBodyProblem::runTwoBodyCalculationWithRenderer(double starMass,
		double planetMass, double e, double a, double startTheta, float dt) {

	TheTwoBodyProblem *twoBody = new TheTwoBodyProblem(starMass, planetMass, dt);
	twoBody->setOrbit(e, a);
	twoBody->setStartPosition(startTheta);
	twoBody->printOrbitalInfo();
	twoBody->dt = dt;

	SimpleRenderer *renderer = new SimpleRenderer(twoBody, dt, false);
	renderer->setupAndRun(800, 800);

	delete renderer;

}

/* Initialization method */

TheTwoBodyProblem* TheTwoBodyProblem::initializeTwoBodySystem(double starMass,
		double planetMass, double e, double a, double startTheta, float dt) {

	TheTwoBodyProblem *twoBody = new TheTwoBodyProblem(starMass, planetMass, dt);
	twoBody->setOrbit(e, a);
	twoBody->setStartPosition(startTheta);
	twoBody->errorOverTime = {0};
	twoBody->dt = dt;

	return twoBody;

}

/* Analytical Value calculation Functions */

void TheTwoBodyProblem::moveOrbitingPlanetAhead(double dt, bool useLog, float3 simulationPos, int body_i) {

	if (a == -1 || e == -1) {
		printf("Error: Orbit parameters not set\n");
		printf("Call the setOrbit() function\n");
		return;
	}

	if ((this->t - this->t_peri) >= this->T) {

		numOrbits++;
		this->t_peri = this->t - ((this->t - this->t_peri) - this->T);
		printSimulationError(simulationPos, body_i);

	}

	//Find Mean Anomaly
	double M = calculateMeanAnomality(n, t, t_peri);

	//Set start guess for Eccentric Anomality
	double E_0 = M + (((0 < sin(M)) - (sin(M) < 0))) * 0.85 * e;

	//Find approximate Eccentric Anomality
	double E = approximateEccentricAnomaly(this->e, M, E_0);

	//Semi-Minor Axis b
	double b = sqrt(this->a * this->a * (1 - this->e * this->e));

	//New position
	planetPos.x = this->a * (cos(E) - this->e);
	planetPos.y = b * sin(E);

	//Update values
	this->t += dt;

}

/* Initial Value setters */

void TheTwoBodyProblem::setStartPosition(double theta) {

	if (a == -1 || e == -1) {
		printf("Error: Orbit parameters not set\n");
		printf("Call the setOrbit() function\n");
		return;
	}

	//Find Start Distance
	double r = this->a * (1 - this->e * this->e) / (1 + this->e * cos(theta));
	this->r_0 = r;

	planetStartPosition.x = r * cos(theta);
	planetStartPosition.y = r * sin(theta);
	planetStartPosition.z = 0;

	planetPos.x = planetStartPosition.x;
	planetPos.y = planetStartPosition.y;

	//Also set t_peri based on the initial position
	this->t_peri = (theta / (2 * M_PI)) * this->T;

	//Find Start velocity

	planetStartVelocity.x = (-1)
			* (this->n * this->a / sqrt(1 - this->e * this->e)) * sin(theta);
	planetStartVelocity.y = (this->n * this->a / sqrt(1 - this->e * this->e))
			* (e + cos(theta));
	planetStartVelocity.z = 0;

}

void TheTwoBodyProblem::setOrbit(double e, double a) {

	this->e = e;
	this->a = a;

	//Find Phi
	double phi = G * (starMass + planetMass);

	//Calculate Orbital Period
	this->T = calculateOrbitalPeriod(this->a, phi);

	//Find Mean motion
	this->n = calculateMeanMotion(T);

}

/*Constructor and deconstructor */

TheTwoBodyProblem::TheTwoBodyProblem(double starMass, double planetMass, float dt) {

	this->starMass = starMass;
	this->planetMass = planetMass;
	this->e = -1;
	this->a = -1;
	r_0 = 0;
	t = 0;
	t_peri = 0;
	n = 0;
	T = 0;
	numOrbits = 0;
	this->dt = dt;

}

/* Math Functions */

//Approximated using the Newton-Raphson method
double TheTwoBodyProblem::approximateEccentricAnomaly(double e, double M,
		double E_0) {

	int max_iterations = 100;
	double threshold = 0.00001;

	double E_last = E_0;
	double E_current = E_last
			- ((E_last - e * sin(E_last) - M) / (1 - e * cos(E_last)));

	int i = 0;
	while (abs(E_last - E_current) > threshold && i < max_iterations) {
		i += 1;

		E_last = E_current;
		E_current = E_last
				- ((E_last - e * sin(E_last) - M) / (1 - e * cos(E_last)));

	}

	if (i >= max_iterations) {
		printf("Warning: Eccentric Anomaly did not converge\n");
	}

	return E_current;

}

double TheTwoBodyProblem::calculateMeanAnomality(double n, double t,
		double t_peri) {

	return n * (t - t_peri);

}

double TheTwoBodyProblem::calculateMeanMotion(double T) {

	return 2 * M_PI / T;

}
double TheTwoBodyProblem::calculateOrbitalPeriod(double a, double phi) {

	return sqrt((4 * M_PI * M_PI / phi) * a * a * a);

}

/* Getters and setters */

double TheTwoBodyProblem::getA() {

	return this->a;

}

double TheTwoBodyProblem::getEccentricity() {

	return this->e;

}

double TheTwoBodyProblem::getMeanMotion() {

	return this->n;
}

double3 TheTwoBodyProblem::getPlanetStartPosition() {

	return planetStartPosition;

}

double3 TheTwoBodyProblem::getPlanetStartVelocity() {

	return planetStartVelocity;

}

double3 TheTwoBodyProblem::getPlanetPosition() {

	return planetPos;

}

double TheTwoBodyProblem::getPlanetMass() {

	return planetMass;

}

double TheTwoBodyProblem::getStarMass() {

	return starMass;

}

double TheTwoBodyProblem::getTimestep() {

	return dt;
}

void TheTwoBodyProblem::setTimestep(double dt) {

	this->dt = dt;
}

void TheTwoBodyProblem::printOrbitalInfo() {

	double perihelion = a * (1 - e);
	double aphelion = a * (1 + e);
	double b = sqrt(a * a * (1 - e * e));
	double peri_v = n * a * sqrt((1 + e) / (1 - e));
	double aphe_v = n * a * sqrt((1 - e) / (1 + e));

	printf("Orbital Information\n");
	printf("-----------------------\n");
	printf("a: %f, b: %f, e: %f, n: %f, T: %f, r_0: %f\n", a, b, e, n, T, r_0);
	printf("Planet Position(X,Y): %f,%f, with t_peri: %f\n", planetPos.x,
			planetPos.y, t_peri);
	printf("Planet Start Velocity(X,Y): %f, %f \n", planetStartVelocity.x,
			planetStartVelocity.y);
	printf("Velocity perihelion: %f, aphelion: %f\n", peri_v, aphe_v);
	printf("Perihelion: %f, Aphelion: %f\n", perihelion, aphelion);
	printf("-----------------------\n");
	printf("\n");
}

std::vector<float> TheTwoBodyProblem::getErrorOverTime() {

	return errorOverTime;
}

void TheTwoBodyProblem::printSimulationError(float3 simulationPos, int body_i) {

	float3 err;
	err.x = simulationPos.x - planetPos.x;
	err.y = simulationPos.y - planetPos.y;
	err.z = simulationPos.z - planetPos.z;

	float abs_err = sqrt(err.x * err.x + err.y * err.y + err.z * err.z);

	this->errorOverTime.push_back(abs_err);

	printf("BODY %d(Orbit %d) Error at time t = %f: X: %f, Y: %f, Z: %f Abs: %f\n", body_i, numOrbits,t, err.x,
			err.y, err.z, abs_err);

}

