#include "CelestialNbody.h"
#include "NbodyInteractions.h"
#include "../utility/CudaErrorCheck.h"
#include "../input/UserInput.h"
#include <stdio.h>

//CPU
bool CelestialNbody::celestialNbodyCPU(float3 *acceleration, float3 *position, float *mass, int solarSystems, int solarBodies, bool enableThresholdSwitching) {

	if (UserInput::logSimulation) {
		printf("[Log] Simulation Algorithm: Celestial N-Body CPU Algorithm\n");
	}

	float4 *massCenters = calculateInternalAccelerationCPU(acceleration, position, mass, solarSystems, solarBodies);
	bool thresholdReached = calculateExternalAccelerationCPU(acceleration, position, mass, massCenters, solarSystems, solarBodies, true, UserInput::distanceThreshold);

	free(massCenters);

	return enableThresholdSwitching && thresholdReached;

}

float4 *CelestialNbody::calculateInternalAccelerationCPU(float3 *acceleration, float3 *position, float *mass, int solarSystems, int solarBodies) {

	float4 *massCenters = (float4*) calloc(solarSystems, sizeof(float4));

	for (int solarIndex=0;solarIndex<solarSystems;solarIndex++) {

		int start = (solarIndex * solarBodies) + 1;
		int end = start + solarBodies;

		float totalMass = 0.0;
		float xSum = 0.0;
		float ySum = 0.0;
		float zSum = 0.0;

		for (int i=start;i<end;i++) {

			float3 i_acc = make_float3(0,0,0);

			for (int j=start;j<end;j++) {

				if (i == j) {
					continue;
				}

				i_acc = computeBodyBody(position[i], mass[i], position[j], mass[j], i_acc);

			}

			acceleration[i] = i_acc;

			totalMass += mass[i];
			xSum +=	mass[i] * position[i].x;
			ySum += mass[i] * position[i].y;
			zSum += mass[i] * position[i].z;
		}

		massCenters[solarIndex] = make_float4(xSum / totalMass, ySum / totalMass, zSum / totalMass, totalMass);

	}

	return massCenters;

}

bool CelestialNbody::calculateExternalAccelerationCPU(float3 *acceleration, float3 *position, float *mass, float4 *massCenters,
		int solarSystems, int solarBodies, bool sumAcc, float distanceThreshold) {

	int N = 1 + solarBodies * solarSystems;

	bool thresholdReached = false;

	for (int i=1;i<N;i++) {

		int solarIndex = (i - 1) / solarBodies;

		float3 i_acc;
		if (sumAcc) {
			i_acc = acceleration[i];
		} else {
			i_acc = make_float3(0,0,0);
		}

		i_acc = computeBodyBody(position[i], mass[i], position[0], mass[0], i_acc);

		for (int j=0;j<solarSystems;j++) {

			float3 j_pos = make_float3(massCenters[j].x, massCenters[j].y, massCenters[j].z);
			float j_mass = massCenters[j].w;

			if (j != solarIndex) {
				i_acc = computeBodyBody(position[i], mass[i], j_pos, j_mass, i_acc);

				if (distanceThreshold > 0) { //This can be optimized by putting this code in a separate loop.

					float3 i_com = make_float3(massCenters[solarIndex].x,
							massCenters[solarIndex].y, massCenters[solarIndex].z);

					float3 distanceVector = calculateDistanceVector(i_com, j_pos);
					float absoluteDistance = calculateAbsoluteDistance(distanceVector);

					if (absoluteDistance <= distanceThreshold) {
						thresholdReached = true;
					}

				}
			}

		}

		acceleration[i] = i_acc;

	}

	return thresholdReached;

}

bool CelestialNbody::launchCelestialNbodyKernel(float3 *acceleration_d, float3 *position_d, float *mass_d, int solarSystems, int solarBodies, bool enableThresholdSwitching) {

	if (UserInput::logSimulation) {
		printf("[Log] Simulation Algorithm: Celestial N-Body GPU Algorithm\n");
	}

	int N = 1 + solarBodies * solarSystems;

	int block_size = solarBodies;
	int num_blocks = solarSystems;
	int sharedMemSize = block_size * sizeof(float3) + block_size * sizeof(float);

	//Allocate memory in cuda kernel
	float4 *massCenters;
	CudaSafeCall( cudaMalloc(&massCenters, sizeof(float) * 4 * solarSystems) );

	bool *thresholdReachedGPU;
	CudaSafeCall( cudaMalloc(&thresholdReachedGPU, sizeof(bool)) );

	//Step 1: Calculate Internal Accelerations
	calculateInternalAccelerationGPU<<<num_blocks, block_size, sharedMemSize>>>(
			acceleration_d, position_d, mass_d, massCenters, N);
	CudaCheckError();

	int ext_blockSize = UserInput::allPairsBlockSize;
	int ext_numBlocks = (N-1) / ext_blockSize + (((N-1) % ext_blockSize == 0) ? 0 : 1);

	//Step 2: Calculate External Accelerations
	calculateExternalAccelerationGPU<<<ext_numBlocks, ext_blockSize>>>(acceleration_d, position_d, mass_d,
			massCenters, solarSystems, solarBodies, true, UserInput::distanceThreshold, thresholdReachedGPU);
	CudaCheckError();

	bool thresholdReached_output = false;
	if (enableThresholdSwitching) {
		bool *thresholdReached = (bool*) malloc(sizeof(bool));
		CudaSafeCall(
				cudaMemcpy(thresholdReached, thresholdReachedGPU, sizeof(bool),
						cudaMemcpyDeviceToHost));
		thresholdReached_output = *thresholdReached;
	}

	//Free Memory
	CudaSafeCall( cudaFree(massCenters) );
	CudaSafeCall( cudaFree(thresholdReachedGPU) );

	return thresholdReached_output;

}

//GPU Kernels

__global__ void calculateInternalAccelerationGPU(float3 *acceleration_d, float3 *position_d, float *mass_d, float4 *output_com, int N) {

	//Find the index of this planet in the array
	int planetIndex = blockIdx.x * blockDim.x + threadIdx.x + 1;
	int solarGlobalIndex = blockIdx.x;
	int solarBodies = blockDim.x;

	//Define shared memory
	extern __shared__ float sharedMemory[];

	//Divide into float3 position and float mass
	float *massShared = sharedMemory;
	float3 *positionShared = (float3*) &massShared[solarBodies];

	float i_mass;
	float3 i_pos;

	if (planetIndex < N) {

		//Load position values for body i
		i_pos = position_d[planetIndex];

		//Load mass value for body i
		i_mass = mass_d[planetIndex];

		//Load data into the block's shared memory
		positionShared[threadIdx.x] = position_d[planetIndex];
		massShared[threadIdx.x] = mass_d[planetIndex];

	}

	__syncthreads();

	float3 i_acc = make_float3(0,0,0);

	if (planetIndex < N) {

		float3 j_pos;
		float j_mass;

		float totalMass = 0.0;
		float xSum = 0.0;
		float ySum = 0.0;
		float zSum = 0.0;

		for (int j = 0; j < solarBodies; j++) {

			j_pos = positionShared[j];
			j_mass = massShared[j];

			i_acc = computeBodyBody(i_pos, i_mass, j_pos, j_mass, i_acc);

			if (output_com) {
				totalMass += j_mass;
				xSum +=	j_mass * j_pos.x;
				ySum += j_mass * j_pos.y;
				zSum += j_mass * j_pos.z;
			}

		}

		if (threadIdx.x == 0 && output_com) {
			output_com[solarGlobalIndex] = make_float4(xSum / totalMass, ySum / totalMass, zSum / totalMass,
					totalMass);
		}

	}

	__syncthreads();

	if (planetIndex < N) {
		acceleration_d[planetIndex] = i_acc;
	}


}

