#include "CelestialNbodyGrouped.h"
#include "NbodyInteractions.h"
#include "../utility/CudaErrorCheck.h"
#include "../input/UserInput.h"

//Kernel Launcher
bool CelestialNbodyGrouped::launchCelestialNbodyKernel(float3 *acceleration_d, float3 *position_d, float *mass_d, int solarSystems, int solarBodies, bool enableThresholdSwitching) {

	if (UserInput::logSimulation) {
		printf("[Log] Simulation Algorithm: Celestial N-Body GPU Grouped Algorithm\n");
	}

	int N = 1 + solarBodies * solarSystems;

	int block_size = UserInput::celestialNbodyBlockSize;
	int num_blocks = (N - 1) / block_size + (((N - 1) % block_size == 0) ? 0 : 1);
	int systems_per_block = block_size / solarBodies;
	int sharedMemSize = systems_per_block * solarBodies * sizeof(float3) + solarBodies * systems_per_block * sizeof(float);

	//Allocate memory in cuda kernel
	float4 *massCenters;
	CudaSafeCall( cudaMalloc(&massCenters, sizeof(float) * 4 * solarSystems) );

	bool *thresholdReachedGPU;
	CudaSafeCall( cudaMalloc(&thresholdReachedGPU, sizeof(bool)) );

	//Step 1: Calculate Internal Accelerations
	calculateInternalAccelerationGroupedGPU<<<num_blocks, block_size, sharedMemSize>>>(
			acceleration_d, position_d, mass_d, massCenters, solarSystems, solarBodies, systems_per_block);
	CudaCheckError();

	int ext_blockSize = UserInput::allPairsBlockSize;
	int ext_numBlocks = (N-1) / ext_blockSize + (((N-1) % ext_blockSize == 0) ? 0 : 1);

	//Step 2: Calculate External Accelerations
	calculateExternalAccelerationGPU<<<ext_numBlocks, ext_blockSize>>>(acceleration_d, position_d,
			mass_d, massCenters, solarSystems, solarBodies, true, UserInput::distanceThreshold, thresholdReachedGPU);
	CudaCheckError();

	bool thresholdReached_output = false;
	if (enableThresholdSwitching) {
		bool *thresholdReached = (bool*) malloc(sizeof(bool));
		CudaSafeCall(
				cudaMemcpy(thresholdReached, thresholdReachedGPU, sizeof(bool),
						cudaMemcpyDeviceToHost));
		thresholdReached_output = *thresholdReached;
	}

	//Free Memory
	CudaSafeCall( cudaFree(massCenters) );
	CudaSafeCall( cudaFree(thresholdReachedGPU) );

	return thresholdReached_output;

}

__global__ void calculateInternalAccelerationGroupedGPU(float3 *acceleration_d, float3 *position_d, float *mass_d, float4 *output_com, int solarSystems, int solarBodies, int SystemsPerBlock) {

	//Total Size of array
	int N = 1 + solarSystems * solarBodies;

	//Group size
	int groupSize = SystemsPerBlock * solarBodies;

	//Find Indices
	int globalIndex = blockIdx.x * blockDim.x + threadIdx.x + 1; //Global Index in array from 1 to N-1
	int localIndex = threadIdx.x; //Index within group of solar systems from 0 to systemsPerBlock * solarBodies - 1
	int solarGlobalIndex = (globalIndex - 1) / solarBodies; //Index within solar systems from 0 to solarSystems-1
	int solarLocalIndex = localIndex / solarBodies; //Solar system index within current block

	//Define shared memory
	extern __shared__ float sharedMemory[];

	//Divide into float3 position and float mass
	float *massShared = sharedMemory;
	float3 *positionShared = (float3*) &massShared[groupSize];

	float i_mass;
	float3 i_pos;

	if (localIndex < groupSize && globalIndex < N) {

		//Load position values for body i
		i_pos = position_d[globalIndex];

		//Load mass value for body i
		i_mass = mass_d[globalIndex];

		//Load data into the block's shared memory
		positionShared[threadIdx.x] = position_d[globalIndex];
		massShared[threadIdx.x] = mass_d[globalIndex];

	}

	__syncthreads();

	float3 i_acc = make_float3(0,0,0);

	if (localIndex < groupSize && globalIndex < N) {

		float3 j_pos;
		float j_mass;

		float totalMass = 0.0;
		float xSum = 0.0;
		float ySum = 0.0;
		float zSum = 0.0;


		int j_start = solarLocalIndex * solarBodies; //Shared Memory Index
		int j_end = j_start + solarBodies;

		for (int j = j_start; j < j_end; j++) {

			j_pos = positionShared[j];
			j_mass = massShared[j];

			i_acc = computeBodyBody(i_pos, i_mass, j_pos, j_mass, i_acc);

			//Calculate components for Center of mass
			if (output_com) {
				totalMass += j_mass;
				xSum +=	j_mass * j_pos.x;
				ySum += j_mass * j_pos.y;
				zSum += j_mass * j_pos.z;
			}

		}

		if (threadIdx.x % solarBodies == 0 && output_com) {
			output_com[solarGlobalIndex] = make_float4(xSum / totalMass, ySum / totalMass, zSum / totalMass,
					totalMass);
		}

	}

	__syncthreads();

	if (localIndex < groupSize && globalIndex < N) {
		acceleration_d[globalIndex] = i_acc;
	}

}

