#include "NbodyAllPairs.h"
#include "NbodyInteractions.h"
#include "../input/UserInput.h"

//Based on N-Body algorithm developed by NVIDIA in GPU Gems 3
__device__ float3 compute_tile(float3 i_pos, float i_mass, float3 i_acc,
		int N, int threadID) {

	//Fetch the shared memory pointer
	extern __shared__ float sharedMemory[];

	//Divide into float3 position and float mass
	float *massShared = sharedMemory;
	float3 *positionShared = (float3*) &massShared[blockDim.x];

	//Serial for each column in tile
	for (int j = 0; j < blockDim.x; j++) {
		if (threadID < N) {
			i_acc = computeBodyBody(i_pos, i_mass, positionShared[j], massShared[j],
					i_acc);
		}
	}

	return i_acc;

}

//Based on N-Body algorithm developed by NVIDIA in GPU Gems 3
__global__ void perform_nbody_calculation(float3 *d_acceleration,
		float3 *d_position, float *d_mass, int N) {

	//Find threadID
	int globalThreadID = blockIdx.x * blockDim.x + threadIdx.x;

	//Define shared memory
	extern __shared__ float sharedMemory[];

	//Divide into float3 position and float mass
	float *massShared;
	float3 *positionShared;
	massShared = sharedMemory;
	positionShared = (float3*) &massShared[blockDim.x];

	float3 i_acc = make_float3(0,0,0);
	float3 i_pos = make_float3(0,0,0);
	float i_mass;

	if (globalThreadID < N) {

		//Load position values for Body i
		i_pos = d_position[globalThreadID];

		//Load mass of body i
		i_mass = d_mass[globalThreadID];

	}

	int j, tile;
	for (j = 0, tile = 0; j < N; j += blockDim.x, tile++) {

		//Find Thread ID
		int threadID = tile * blockDim.x + threadIdx.x;

		//Copy information for Body i.
		if (threadID < N) {
			positionShared[threadIdx.x] = d_position[threadID];
			massShared[threadIdx.x] = d_mass[threadID];
		}

		//Sync threads
		__syncthreads();

		i_acc = compute_tile(i_pos, i_mass, i_acc, N, threadID);

		//Sync threads
		__syncthreads();

	}

	//Save the new acceleration to global memory
	if (globalThreadID < N) {
		d_acceleration[globalThreadID] = i_acc;
	}

}

//All Pairs Nbody CPU version to compare with the GPU version
void nbody_allpairs_cpu(float3 *acceleration, float3 *position, float *mass,
		int N) {

	if (UserInput::logSimulation) {
		printf("[Log] Simulation Algorithm: All-Pairs CPU N-Body Algorithm \n");
	}

	//Future Work?: Create logic to re-enable Celestial N-Body CPU?

	//Perform computation
	for (int i = 0; i < N; i++) {

		//Set initial values for each timestep
		float3 i_pos = position[i];
		float i_mass = mass[i];
		float3 i_acc = make_float3(0,0,0);

		for (int j = 0; j < N; j++) {
			if (i == j) {
				continue;
			}

			float3 j_pos = position[j];
			float j_mass = mass[j];

			//Find the new acceleration from the N bodies
			i_acc = computeBodyBody(i_pos, i_mass, j_pos, j_mass, i_acc);
		}

		//Copy the value back to the array
		acceleration[i] = i_acc;

	}

}

//Based on N-Body algorithm developed by NVIDIA in GPU Gems 3
void nbody_allpairs_gpu(float3 *d_acceleration, float3 *d_position,
		float *d_mass, int N) {

	if (UserInput::logSimulation) {
		printf("[Log] Simulation Algorithm: All-Pairs GPU N-Body Algorithm \n");
	}

	//Future Work?: Create logic to re-enable Celestial N-Body GPU?

	//Launch Configuration
	int block_size = UserInput::allPairsBlockSize;
	int num_blocks = N / block_size + ((N % block_size == 0) ? 0 : 1);
	int sharedMemSize = block_size * sizeof(float3)
			+ block_size * sizeof(float);

	perform_nbody_calculation<<<num_blocks, block_size, sharedMemSize>>>(
			d_acceleration, d_position, d_mass, N);
	CudaCheckError();

}
