#include "NbodyInteractions.h"
#include "../input/CelestialConstants.h"
#include <stdio.h>

__host__ __device__ float3 computeBodyBody(float3 i_pos, float i_mass, float3 j_pos, float j_mass, float3 i_acc) {

	float3 distanceVector = calculateDistanceVector(i_pos, j_pos);

	//Don't do anything if the two bodies are at the exact same position
	if (distanceIsZero(distanceVector)) {
		return i_acc;
	}

	float absoluteDistance = calculateAbsoluteDistance(distanceVector);

	float absoluteDistanceCubed = absoluteDistance * absoluteDistance * absoluteDistance;

	float3 dA = calculateAccelerationChange(j_mass, absoluteDistanceCubed, distanceVector);

	i_acc = sumAccelerationChange(i_acc, dA);

	return i_acc;

}

//Debug function - Used to test different order of computation for GPU and/or CUDA intrinsics
//Based on N-Body algorithm developed by NVIDIA in GPU Gems 3
__device__ float3 computeBodyBodyGPU(float3 i_pos, float i_mass, float3 j_pos, float j_mass, float3 i_acc) {

	float3 r = calculateDistanceVector(i_pos, j_pos);

	//Don't do anything if the two bodies are at the exact same position
	if (distanceIsZero(r)) {
		return i_acc;
	}

	float r_squared = r.x * r.x + r.y * r.y + r.z * r.z + relaxationFactor;
	float r_sixth = r_squared * r_squared * r_squared;
	float invR_cubed = 1.0f / sqrtf(r_sixth);
	float s = j_mass * invR_cubed;

	i_acc.x += r.x * s;
	i_acc.y += r.y * s;
	i_acc.z += r.z * s;

	return i_acc;

}


__host__ __device__ bool distanceIsZero(float3 distanceVector) {

	return (distanceVector.x == 0 && distanceVector.y == 0 && distanceVector.z == 0);

}

__host__ __device__ float3 calculateDistanceVector(float3 pos_i, float3 pos_j) {

	float3 distanceVector;
	distanceVector.x = pos_j.x - pos_i.x;
	distanceVector.y = pos_j.y - pos_i.y;
	distanceVector.z = pos_j.z - pos_i.z;

	return distanceVector;

}

__host__ __device__ float calculateAbsoluteDistance(float3 distanceVector) {

	return sqrtf(distanceVector.x * distanceVector.x + distanceVector.y * distanceVector.y +
		   distanceVector.z * distanceVector.z + relaxationFactor);
}

__host__ __device__ float3 calculateAccelerationChange(float j_mass, float absoluteDistanceCubed,
		float3 distanceVector) {

	float3 dA;
	dA.x = (G * j_mass / absoluteDistanceCubed) * distanceVector.x;
	dA.y = (G * j_mass / absoluteDistanceCubed) * distanceVector.y;
	dA.z = (G * j_mass / absoluteDistanceCubed) * distanceVector.z;

	return dA;

}

__host__ __device__ float3 sumAccelerationChange(float3 i_acc, float3 dA) {

	float3 newAcceleration;
	newAcceleration.x = i_acc.x + dA.x;
	newAcceleration.y = i_acc.y + dA.y;
	newAcceleration.z = i_acc.z + dA.z;

	return newAcceleration;

}

__host__ __device__ float4 calculateMassCenter(float3 *position, float *mass, int startIndex, int endIndex) {

	float totalMass = 0.0;
	float xSum = 0.0;
	float ySum = 0.0;
	float zSum = 0.0;

	for (int i=startIndex; i < endIndex; i++) {

		totalMass += mass[i];
		xSum += mass[i] * position[i].x;
		ySum += mass[i] * position[i].y;
		zSum += mass[i] * position[i].z;

	}

	float4 massCenter = make_float4(xSum / totalMass, ySum / totalMass, zSum / totalMass,
				totalMass);

	return massCenter;

}

__global__ void calculateMassCenterForSolarSystems(float3 *position, float *mass, float4 *output_com) {

	int globalID = blockIdx.x * blockDim.x + threadIdx.x + 1;
	int start = globalID;
	int end = start + blockDim.x;
	int solarGlobalIndex = (globalID - 1) / blockDim.x;

	if (threadIdx.x == 0) {
		output_com[solarGlobalIndex] = calculateMassCenter(position, mass, start, end);
	}

}

float4 *calculateMassCenterForSolarSystemsCPU(float3 *position, float *mass, int solarSystems, int solarBodies) {

	float4 *massCenters = (float4*) calloc(solarSystems, sizeof(float4));

	for (int i=0;i<solarSystems;i++) {

		int start = (i * solarBodies) + 1;
		int end = start + solarBodies;

		massCenters[i] = calculateMassCenter(position, mass, start, end);

		/*
		printf("Mass Center Celestial Integration CPU %d\n", i);
		printf("X: %f Y: %f Z: %f W: %f\n", massCenters[i].x, massCenters[i].y, massCenters[i].z , massCenters[i].w);
		printf("\n");
		*/


	}

	return massCenters;

}


__global__ void calculateExternalAccelerationGPU(float3 *acceleration_d, float3 *position_d, float *mass_d, float4 *massCenters,
		int solarSystems, int solarBodies, bool sumAcc, float distanceThreshold, bool *thresholdReached) {

	int globalIndex = blockIdx.x * blockDim.x + threadIdx.x + 1;
	int N = 1 + solarSystems * solarBodies;
	int solarGlobalIndex = (globalIndex - 1) / solarBodies;

	float3 i_acc; //returned value

	if (globalIndex < N) {

		float3 i_pos = position_d[globalIndex];
		float i_mass = mass_d[globalIndex];

		if (sumAcc) {
			i_acc = acceleration_d[globalIndex];
		} else {
			i_acc = make_float3(0,0,0);
		}

		//The acceleration from the black hole
		i_acc = computeBodyBody(i_pos, i_mass, position_d[0], mass_d[0], i_acc);

		//The acceleration from each distant solar system
		float3 j_pos;
		float j_mass;


		if (distanceThreshold > 0 && globalIndex == 1) {
			*thresholdReached = false;
		}

		for (int j = 0; j < solarSystems; j++) {

			if (j != solarGlobalIndex) {
				j_pos = make_float3(massCenters[j].x, massCenters[j].y, massCenters[j].z);
				j_mass = massCenters[j].w;

				i_acc = computeBodyBody(i_pos, i_mass, j_pos, j_mass, i_acc);

				if (distanceThreshold > 0 && globalIndex == 1) {

					float3 i_com = make_float3(massCenters[solarGlobalIndex].x,
							massCenters[solarGlobalIndex].y, massCenters[solarGlobalIndex].z);

					float3 distanceVector = calculateDistanceVector(i_com, j_pos);
					float absoluteDistance = calculateAbsoluteDistance(distanceVector);

					if (absoluteDistance <= distanceThreshold) {
						*thresholdReached = true;
					}

				}

			}

		}

	}

	__syncthreads();

	if (globalIndex < N) {
		acceleration_d[globalIndex] = i_acc;
	}

}

