#include "CelestialRenderer.h"
#include "stdio.h"
#include "../math/rendering/CelestialStructs.h"
#include "../utility/OpenglUtility.h"
#include "../utility/CudaErrorCheck.h"
#include "../input/UserInput.h"
#include "../utility/Utilities.h"


#define ToRadian(x) ((x) * M_PI / 180.0f)
#define ToDegree(x) ((x) * 180.0f / M_PI)
#define ROTATE_STEP 10
#define ZOOM_STEP 10

CelestialRenderer::CelestialRenderer(CelestialSystem *renderingSystem, float dt) {

	this->renderingSystem = renderingSystem;
	this->dt = dt;
	shouldQuit = false;
	window = 0;
	renderer = 0;
	fov = UserInput::blackHoleClosestDistance * 3 + UserInput::galaxyPlanesDistance * UserInput::galaxyPlanes;
	screen_height = 0;
	screen_width = 0;
	camDistance = 1;
	camVertical = 0;
	camHorizontal = 40.00;
	cameraTarget = PVector3f(0, 0, 0);
	cameraOrientation = PVector3f(0, 1, 0);
	runSimulation = false;
	pipeline = 0;
	shaderProgram = 0;
	gWorldID = 0;
	glContext = 0;
	dataRecordInterval = 365 * 10;
	timeSinceLastDataRecord = 0;

}

void CelestialRenderer::renderBodies() {

	renderingSystem->drawBodies();

}

void CelestialRenderer::prepareToRender() {

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glClearColor(0.0f, 0.0f, 0.2f, 0.0f);

	glUniformMatrix4fv(gWorldID, 1, GL_TRUE,
			(const GLfloat *) pipeline->getTransformation(cameraPos,
					cameraTarget, cameraOrientation));

}

void CelestialRenderer::moveBodies() {

	if (runSimulation) {
		renderingSystem->moveBodies(dt);
	}

	//Do necessary checks
	if (UserInput::pauseSimulationAtTimestep != 0 && renderingSystem->getCurrentTime() > UserInput::pauseSimulationAtTimestep) {
		printf("[Simulation Output] Paused at time %.f days\n", renderingSystem->getCurrentTime());
		UserInput::pauseSimulationAtTimestep = 0;
		runSimulation = false;
	}

}

void CelestialRenderer::setupAndRun(int screenWidth, int screenHeight) {

	this->screen_width = screenWidth;
	this->screen_height = screenHeight;

	//Initialize SDL
	if (SDL_Init(SDL_INIT_VIDEO) != 0) {
		printf("Error initializing SDL\n");
		exit(EXIT_FAILURE);
	}

	//Ask for a OpenGL 3.2 Core Context
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK,
			SDL_GL_CONTEXT_PROFILE_CORE);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 3);

	Uint32 windowFlags = SDL_WINDOW_SHOWN | SDL_WINDOW_OPENGL;

	//Create window
	window = SDL_CreateWindow("CelestialRenderer", 0, 0, screenWidth,
			screenHeight, windowFlags);
	if (!window) {
		printf("Error creating window\n");
		SDL_Quit();
		exit(EXIT_FAILURE);
	}

	//Create OpenGL Context
	glContext = SDL_GL_CreateContext(window);
	if (glContext == NULL)
	{
		fprintf(stderr,"Error creating OpenGL context!\n");
		exit(EXIT_FAILURE);
	}

	glewExperimental = GL_TRUE;

	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
	  /* Problem: glewInit failed, something is seriously wrong. */
	  fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
	  exit(EXIT_FAILURE);
	}

	//Set Swap interval
	SDL_GL_SetSwapInterval(1);

	//Setup Transformation Pipeline
	pipeline = new TransformationPipeline();

	//Set Orthographic Projection
	setProjection();

	//Create Shader Program
	shaderProgram = OpenglUtility::buildProgram();
	gWorldID = glGetUniformLocation(shaderProgram, "gWorld");

	glUseProgram(shaderProgram);

	//Setup the system for rendering
	renderingSystem->setupForRendering();

	//Set Camera
	setCameraOrbitPosition();

	runRenderingLoop();

}

void CelestialRenderer::runRenderingLoop() {

	while (!shouldQuit) {

		handleInput();

		prepareToRender();

		renderBodies();

		moveBodies();

		SDL_GL_SwapWindow(window);

	}

	SDL_GL_DeleteContext(glContext);

	SDL_DestroyWindow(window);

	SDL_Quit();

}

void CelestialRenderer::graphMovement() { //Unused

	int N = renderingSystem->getN();

	float *x = (float*) malloc(sizeof(float) * positionData.size() * N);
	float *y = (float*) malloc(sizeof(float) * positionData.size() * N);
	float *z = (float*) malloc(sizeof(float) * positionData.size() * N);

	for(int i=0;i<positionData.size();i++) {

		float3* position = positionData[i];

		for(int j=0;j<N;j++) {

			x[i * N + j] = position[j].x;
			y[i * N + j] = position[j].y;
			z[i * N + j] = position[j].z;

		}

	}

	Utilities::runPython3DScatterPlot(x, y, z, N * positionData.size());

	free(x);
	free(y);
	free(z);

}

void CelestialRenderer::handleInput() {

	SDL_Event event;
	while (SDL_PollEvent(&event)) {

		if (event.key.keysym.sym == SDLK_ESCAPE) {
			shouldQuit = true;
		}

		if (event.type == SDL_QUIT) {
			shouldQuit = true;
		}

		if (event.key.keysym.sym == SDLK_UP) {
			zoomCamera(-ZOOM_STEP);
		}

		if (event.key.keysym.sym == SDLK_DOWN) {
			zoomCamera(ZOOM_STEP);
		}

		if (event.key.keysym.sym == SDLK_LEFT) {
			moveCameraHorizontally(-ROTATE_STEP);
		}

		if (event.key.keysym.sym == SDLK_RIGHT) {
			moveCameraHorizontally(ROTATE_STEP);
		}

		if (event.key.keysym.sym == SDLK_SPACE && event.type == SDL_KEYUP) {
			toggleRunSimulation();
		}

		if (event.key.keysym.sym == SDLK_0 && event.type == SDL_KEYUP) { //Work in progress
			selectSystem(0);
		}

		if (event.key.keysym.sym == SDLK_r) {
			camHorizontal = 40.00;
			setCameraOrbitPosition();
		}

		if (event.key.keysym.sym == SDLK_TAB && event.type == SDL_KEYUP) {
			moveCameraHorizontally(180);
			setCameraOrbitPosition();
		}

		if (event.key.keysym.sym == SDLK_f && event.type == SDL_KEYUP) {
			setFOV(220);
		}

		if (event.key.keysym.sym == SDLK_c && event.type == SDL_KEYUP) {
			renderingSystem->toggleIntegrationMethod();
		}

		if (event.key.keysym.sym == SDLK_p && event.type == SDL_KEYUP) {
				UserInput::measurePerformance = !UserInput::measurePerformance;
		}

		if (event.key.keysym.sym == SDLK_e && event.type == SDL_KEYUP) {
				renderingSystem->printEnergyValues();
		}

		if (event.key.keysym.sym == SDLK_PERIOD && event.type == SDL_KEYUP) {
			printf("[SIMULATION OUTPUT] Current timestep: %.f years\n", renderingSystem->getCurrentTime() / 365.0);
		}

		if (event.key.keysym.sym == SDLK_l && event.type == SDL_KEYUP) {
				UserInput::logSimulation = !UserInput::logSimulation;
				printf("[KEYBOARD INPUT] Log %s\n", UserInput::logSimulation ? "Enabled" : "Disabled");
		}

		if (event.key.keysym.sym == SDLK_v && event.type == SDL_KEYUP) {
			int i = SDL_GL_GetSwapInterval() ^ 1;
			SDL_GL_SetSwapInterval(i);
			printf("[KEYBOARD INPUT] V-Sync %s\n", i == 1 ? "Enabled" : "Disabled");
		}

		if (event.key.keysym.sym == SDLK_a && event.type == SDL_KEYUP) {
			printf("[KEYBOARD INPUT] Current Camera Angle: %f\n", camHorizontal);

		}



	}

}

void CelestialRenderer::setCameraOrbitPosition() {

	float x_horizontal = cameraTarget.x
			+ sin(ToRadian(camHorizontal)) * camDistance;
	float y_horizontal = cameraTarget.z
			+ cos(ToRadian(camHorizontal)) * camDistance;

	PVector3f horizontal = PVector3f(x_horizontal, 0, y_horizontal);

	cameraPos = horizontal;

}

void CelestialRenderer::setProjection() {

	double l = fov / 2.0;
	pipeline->setOrthographicValues(screen_width, screen_height, -l, l, -l, l, -l*1000,
			l*1000);

}

void CelestialRenderer::zoomCamera(float offset) {

	if (!(fov + offset < 0)) {

		fov += offset;

	}

	setProjection();

}

void CelestialRenderer::moveCameraHorizontally(float angleOffset) {

	camHorizontal += angleOffset;

	if (camHorizontal >= 360) {
		camHorizontal = 0;
	}

	setCameraOrbitPosition();

}

void CelestialRenderer::moveCameraVertically(float angleOffset) {

	camVertical += angleOffset;

	if (camVertical >= 360) {
		camVertical = 0;
	}

	setCameraOrbitPosition();

}

void CelestialRenderer::selectSystem(int system_i) { //not used

	float solarBodies = 8;

	float starIndex = (system_i * solarBodies) + 1;

	float3 pos = renderingSystem->getBodyPosition(starIndex);

	cameraTarget = PVector3f(pos.x, pos.y, pos.z);
	cameraPos = PVector3f(pos.x, pos.y, pos.z - 1);
	fov = 50;
	camHorizontal = 0;

	setCameraOrbitPosition();
	setProjection();

}

void CelestialRenderer::setCameraOrientation(PVector3f cameraOrientation) {

	this->cameraOrientation = cameraOrientation;

}

void CelestialRenderer::toggleRunSimulation() {

	runSimulation = !runSimulation;

}

void CelestialRenderer::setCamera(PVector3f pos, PVector3f target,
		PVector3f orientation) {

	this->cameraOrientation = orientation;
	this->cameraPos = pos;
	this->cameraTarget = target;

}

void CelestialRenderer::setFOV(float fov) {

	this->fov = fov;
	setProjection();

}
