#include "SimpleRenderer.h"
#include <cstdlib>
#include <cstdio>
#include <stdio.h>
#include "math.h"
#include "../generator/SystemGenerator.h"
#include "../system/CelestialSystem.h"
#include <string>
#include "../utility/Utilities.h"


SimpleRenderer::SimpleRenderer(CelestialSystem *simulationSystem, float dt) {

	this->simulationSystem = simulationSystem;
	this->dt = dt;
	shouldShowSimulationReport = 0;
	renderSimulation = true;
	renderAnalytical = false;
	shouldQuit = false;
	screen_height = 0;
	screen_width = 0;
	window = 0;

}

SimpleRenderer::SimpleRenderer(TheTwoBodyProblem *twoBody, float dt, bool shouldShowSimulationReport) {

	bodies.push_back(twoBody);

	SystemGenerator *newSystem = SystemGenerator::generateSolarSystem(bodies);
	simulationSystem = new CelestialSystem(newSystem->getPositions(), newSystem->getAccelerations(), newSystem->getVelocities(), newSystem->getMasses(),
			newSystem->getN(), false);
	this->shouldShowSimulationReport = shouldShowSimulationReport;
	renderSimulation = true;
	renderAnalytical = true;
	shouldQuit = false;
	screen_height = 0;
	screen_width = 0;
	window = 0;
	this->dt = dt;

}

SimpleRenderer::SimpleRenderer(std::vector<TheTwoBodyProblem*> bodies, float dt, bool shouldShowSimulationReport) {

	renderSimulation = true;
	renderAnalytical = true;

	SystemGenerator *newSystem = SystemGenerator::generateSolarSystem(bodies);
	simulationSystem = new CelestialSystem(newSystem->getPositions(), newSystem->getAccelerations(), newSystem->getVelocities(), newSystem->getMasses(),
				newSystem->getN(), false);
	this->shouldShowSimulationReport = shouldShowSimulationReport;
	shouldQuit = false;
	screen_height = 0;
	screen_width = 0;
	window = 0;
	this->dt = dt;

	for (int i = 0; i < bodies.size(); i++) {

		this->bodies.push_back(bodies[i]);

	}

}

SimpleRenderer::SimpleRenderer(std::vector<TheTwoBodyProblem*> bodies,
		bool renderSimulation, bool renderAnalytical, float dt, bool shouldShowSimulationReport) {

	this->renderSimulation = renderSimulation;
	this->renderAnalytical = renderAnalytical;
	SystemGenerator *newSystem = SystemGenerator::generateSolarSystem(bodies);
	simulationSystem = new CelestialSystem(newSystem->getPositions(), newSystem->getAccelerations(), newSystem->getVelocities(), newSystem->getMasses(),
				newSystem->getN(), false);
	shouldQuit = false;
	this->shouldShowSimulationReport = shouldShowSimulationReport;
	screen_height = 0;
	screen_width = 0;
	window = 0;
	this->dt = dt;

	for (int i = 0; i < bodies.size(); i++) {
		this->bodies.push_back(bodies[i]);
	}

}

SimpleRenderer::~SimpleRenderer() {

	delete simulationSystem;

}


void SimpleRenderer::setupAndRun(int screenWidth, int screenHeight) {

	this->screen_width = screenWidth;
	this->screen_height = screenHeight;

	//Initialize SDL
	if (SDL_Init(SDL_INIT_EVERYTHING) != 0) {
		printf("Error initializing SDL\n");
		exit(EXIT_FAILURE);
	}

	//Ask for a doublebuffered framebuffer
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	//Create window
	window = SDL_CreateWindow("SimpleRenderer", 0, 0, screenWidth,
			screenHeight, SDL_WINDOW_OPENGL);
	if (!window) {
		printf("Error creating window\n");
		SDL_Quit();
		exit(EXIT_FAILURE);
	}

	//Create OpenGL Context
	SDL_GL_CreateContext (window);

	//Set Swap interval
	SDL_GL_SetSwapInterval(1);

	runRenderingLoop();

}

void SimpleRenderer::runRenderingLoop() {

	while (!shouldQuit) {

		glClearColor(0.0f, 0.0f, 0.2f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT);

		handleInput();

		renderEllipse();

		renderBodies();

		moveBodies();

		SDL_GL_SwapWindow(window);

	}

	SDL_DestroyWindow(window);

	SDL_Quit();

	//Show simulation report
	showReport();

}

void SimpleRenderer::showReport() {

	if (!shouldShowSimulationReport) {
		return;
	}

	float *xArray, *yArray;

	for (int i=0; i<bodies.size();i++) {

		TheTwoBodyProblem *body = bodies[i];

		xArray = (float*) malloc(sizeof(float) * body->getErrorOverTime().size());
		yArray = (float*) malloc(sizeof(float) * body->getErrorOverTime().size());
		float time = 0;
		float steps = body->getErrorOverTime().size();

		for (int j=0; j<body->getErrorOverTime().size(); j++) {

			xArray[j] = time;
			yArray[j] = body->getErrorOverTime()[j];

			time += 1;
		}

		std::string xlabel = "xlabel = 'X axis: Number of orbits dt = " + std::to_string(dt) + " days'";
		std::string ylabel = "ylabel = 'Y axis: The absolute distance difference in AU'";
		std::string title;

		if (bodies.size() == 1) {
			title = "title = 'Two Body Test Case Result for Body " + std::to_string(i) + "'\n";
		} else {
			title = "title = 'Two Body Solar System Test Case Results for Body " + std::to_string(i) + "'\n";
		}

		//Utilities::runPythonPlotScript(xArray, yArray, steps, xlabel.c_str(), ylabel.c_str(), title.c_str());

		printf("-------------------------------------------------\n");
		printf("PyLab Plot Output Body %i (dt = %f)\n", i, dt);
		printf("-------------------------------------------------\n");
		Utilities::printPythonArray("x", xArray, steps);
		Utilities::printPythonArray("y", yArray, steps);
		printf("%s\n", xlabel.c_str());
		printf("%s\n", ylabel.c_str());
		printf("%s\n", title.c_str());
		printf("-------------------------------------------------\n");
		printf("\n");

		free(xArray);
		free(yArray);

	}

}

void SimpleRenderer::handleInput() {

	SDL_Event event;
	while (SDL_PollEvent(&event)) {

		if (event.key.keysym.sym == SDLK_ESCAPE) {
			shouldQuit = true;
		}

		if (event.type == SDL_QUIT) {
			shouldQuit = true;
		}

	}

}

void SimpleRenderer::renderEllipse() {

	if (bodies.size() != 0) {

		double boundary = bodies.back()->getA() * 1.1;
		double offset = 0;

		glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
		glBlendFunc(GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA);
		glEnable(GL_POINT_SMOOTH);

		glColor3f(1, 1, 1);
		glPointSize(1);
		glBegin(GL_POINTS);

		double x, y, r, renderX, renderY;

		for (int it = 0; it < bodies.size(); it++) {

			for (int i = 0; i < 360 * 8; i++) {

				TheTwoBodyProblem *body = bodies[it];

				r = (body->getA()
						* (1
								- body->getEccentricity()
										* body->getEccentricity())
						/ (1 + body->getEccentricity() * cos(i)));
				x = r * cos(i);
				y = r * sin(i);

				renderX = (x + offset) / boundary;
				renderY = (y + offset) / boundary;

				glVertex2d(renderX, renderY);

			}

		}

		glEnd();
	}

}

void SimpleRenderer::renderBodies() {

	double boundary;

	if (bodies.size() != 0) {
		boundary = bodies.back()->getA() * 1.1;
	} else {
		boundary = 110;
	}

	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	glBlendFunc(GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA);
	glEnable(GL_POINT_SMOOTH);

	//Simulation
	if (renderSimulation) {

		//Planets
		glColor3f(1.0, 0.0, 0.0);
		glPointSize(15);
		glBegin(GL_POINTS);
		for (int i = 1; i < simulationSystem->getN(); i++) {
			float3 pos_sim = simulationSystem->getBodyPosition(i);

			glVertex2f(pos_sim.x / boundary, pos_sim.y / boundary);
		}
		glEnd();

		//Sun
		float3 pos_sim_sun = simulationSystem->getBodyPosition(0);
		glColor3f(1.0, 0.0, 0.0);
		glPointSize(30);
		glBegin(GL_POINTS);
		glVertex2f(pos_sim_sun.x / boundary, pos_sim_sun.y / boundary);
		glEnd();

	}

	//Analytical
	if (renderAnalytical) {
		glColor3f(182 / 255.0, 126 / 255.0, 91 / 225.0);
		glPointSize(30);
		glBegin(GL_POINTS);
		glVertex2f(0, 0);
		glEnd();

		glColor3f(182 / 255.0, 126 / 255.0, 91 / 225.0);
		glPointSize(15);
		glBegin(GL_POINTS);
		for (int i = 0; i < bodies.size(); i++) {

			TheTwoBodyProblem *body = bodies[i];

			double3 pos = body->getPlanetPosition();
			glVertex2f(pos.x / boundary, pos.y / boundary);
		}
		glEnd();

	}
}

void SimpleRenderer::moveBodies() {

	for (int i = 0; i < bodies.size(); i++) {
		TheTwoBodyProblem *body = bodies[i];
		body->moveOrbitingPlanetAhead(dt, false, simulationSystem->getBodyPosition(i+1), i);
	}

	//The object at index 0 contains the simulation data
	simulationSystem->moveBodies(dt);

	//The sun is being kept in place
	simulationSystem->setBodyPosition(0,0,0,0);

}

