#include "CelestialSystem.h"
#include <cuda.h>
#include <cstdio>

//Other Class includes
#include "../generator/GalaxyGenerator.h"
#include "../input/UserInput.h"
#include "../input/CelestialConstants.h"
#include "../utility/Utilities.h"
#include "../utility/CudaErrorCheck.h"
#include "../utility/OpenglUtility.h"
#include "../utility/CudaTimer.cuh"
#include "../math/integration/IntegrationModule.h"

/* Integration Method */

void CelestialSystem::performIntegration(float dt) {

	CudaTimer *timer;
	double before, after;
	if (UserInput::measurePerformance) {

		if (useGPU) {
			timer = new CudaTimer();
			timer->cudaTimerStart();
		} else {
			before = Utilities::get_wtime();
		}

	}

	IntegrationModule::performIntegration(this, dt);

	if (UserInput::measurePerformance) {

		if (useGPU) {
			timer->cudaTimerStop();
			timer->printElapsedTime();
			delete timer;
		} else {
			after = Utilities::get_wtime();
			Utilities::printElapsedTime(before, after);
		}

	}

}

void CelestialSystem::moveBodies(float dt) {

	mapResourcesCuda();

	performIntegration(dt);

	unmapResourcesCuda();

	updateCurrentTime(dt);

}

void CelestialSystem::updateBufferCPU() {

	if (renderingEnabled && !useGPU) {
		glBindBuffer(GL_ARRAY_BUFFER, systemVBO);
		glBufferData(GL_ARRAY_BUFFER, 3 * sizeof(float) * getN(), position,
				GL_STREAM_DRAW);
	}

}

void CelestialSystem::mapResourcesCuda() {

	if (renderingEnabled && useGPU) {
		CudaSafeCall(cudaGraphicsMapResources(1, &positionCudaResource, 0));
		CudaSafeCall(
				cudaGraphicsResourceGetMappedPointer((void** ) &position, 0,
						positionCudaResource));
		CudaSafeCall(cudaStreamSynchronize(0));
	}

}

void CelestialSystem::unmapResourcesCuda() {

	if (renderingEnabled && useGPU) {

		CudaSafeCall(cudaGraphicsUnmapResources(1, &positionCudaResource, 0));
		CudaSafeCall(cudaStreamSynchronize(0));

	}

}

/* Rendering Methods */

void CelestialSystem::drawBodies() {

	if (renderingEnabled) {
		updateBufferCPU();
		glBindVertexArray(systemVAO);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);
		glDrawArrays(GL_POINTS, 0, getN());
		glBindVertexArray(0);
	}

}

void CelestialSystem::setupForRendering() {

	if (renderSimulation) {

		glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);

		//Create VAO for the objects (Explicitly needed on Mac)
		GLuint systemVAO;
		glGenVertexArrays(1, &systemVAO);
		glBindVertexArray(systemVAO);
		this->systemVAO = systemVAO;

		//Create VBO for the objects
		GLuint systemVBO;
		glGenBuffers(1, &systemVBO);
		glBindBuffer(GL_ARRAY_BUFFER, systemVBO);
		glBufferData(GL_ARRAY_BUFFER, 3 * sizeof(float) * N, position,
				GL_DYNAMIC_DRAW);
		glEnableVertexAttribArray(0);
		this->systemVBO = systemVBO;
		glBindVertexArray(0);

		//Associate with CUDA
		if (useGPU) {
			CudaSafeCall(
					cudaGraphicsGLRegisterBuffer(&positionCudaResource, systemVBO,
							cudaGraphicsRegisterFlagsNone));
			free(position);
		}

		//Enable Rendering
		renderingEnabled = true;

	}

}

void CelestialSystem::renderingCleanup() {

	if (renderingEnabled) {
		glDeleteBuffers(1, &systemVBO);
		glDeleteBuffers(1, &systemVAO);
		glDisableVertexAttribArray(0);
	}

	renderingEnabled = false;

}

/* Constructor and Deconstructor */

CelestialSystem::CelestialSystem(int solarSystems, int solarBodies,
		bool useGPU) {

	GalaxyGenerator *newSystem = new GalaxyGenerator(solarSystems, solarBodies);

	this->solarSystems = solarSystems;
	this->solarBodies = solarBodies;
	this->useGPU = useGPU;
	this->renderSimulation = true;
	this->integrationMethod = UserInput::integrationMethod;
	this->simulationAlgorithm = UserInput::nbodyAlgorithm;
	this->fallbackIntegrationMethod = IntegrationMethod::SymplecticEuler;
	this->thresholdReached = false;
	this->distanceThreshold = UserInput::distanceThreshold;
	this->N = newSystem->getN();

	if (UserInput::fallbackAlgorithm == NbodyAlgorithm::CelestialNbodyCPU) {
		this->fallbackAlgorithm = NbodyAlgorithm::AllPairsCPU;
	} else if (UserInput::fallbackAlgorithm == NbodyAlgorithm::CelestialNbodyGPU ||
			UserInput::fallbackAlgorithm == NbodyAlgorithm::CelestialNbodyGroupedGPU) {
		this->fallbackAlgorithm = NbodyAlgorithm::AllPairsGPU;
	} else {
		this->fallbackAlgorithm = UserInput::fallbackAlgorithm;
	}

	this->enableThresholdSwitching = UserInput::enableThresholdSwitching &&
			(simulationAlgorithm == NbodyAlgorithm::CelestialNbodyCPU ||
			 simulationAlgorithm == NbodyAlgorithm::CelestialNbodyGPU ||
			 simulationAlgorithm == NbodyAlgorithm::CelestialNbodyGroupedGPU ||
			 integrationMethod == IntegrationMethod::CelestialIntegrationExplicit ||
			 integrationMethod == IntegrationMethod::CelestialIntegrationSymplectic);

	if (useGPU) {

		if (renderSimulation) {
			position = newSystem->getPositions();
		} else {
			position = newSystem->getPositionsGPU();
		}

		velocity = newSystem->getVelocitiesGPU();
		acceleration = newSystem->getAccelerationsGPU();
		mass = newSystem->getMassesGPU();
	} else {

		position = newSystem->getPositions();
		velocity = newSystem->getVelocities();
		acceleration = newSystem->getAccelerations();
		mass = newSystem->getMasses();

	}

	renderingEnabled = false;
	currentTime = 0;
	systemVAO = 0;
	systemVBO = 0;
	positionCudaResource = 0;

}

CelestialSystem::CelestialSystem(int solarSystems, int solarBodies,
		bool useGPU, bool renderSimulation) {

	GalaxyGenerator *newSystem = new GalaxyGenerator(solarSystems, solarBodies);

	this->solarSystems = solarSystems;
	this->solarBodies = solarBodies;
	this->useGPU = useGPU;
	this->renderSimulation = renderSimulation;
	this->integrationMethod = UserInput::integrationMethod;
	this->simulationAlgorithm = UserInput::nbodyAlgorithm;
	this->fallbackIntegrationMethod = IntegrationMethod::SymplecticEuler;
	this->thresholdReached = false;
	this->distanceThreshold = UserInput::distanceThreshold;
	this->N = newSystem->getN();

	if (UserInput::fallbackAlgorithm == NbodyAlgorithm::CelestialNbodyCPU) {
		this->fallbackAlgorithm = NbodyAlgorithm::AllPairsCPU;
	} else if (UserInput::fallbackAlgorithm == NbodyAlgorithm::CelestialNbodyGPU ||
			UserInput::fallbackAlgorithm == NbodyAlgorithm::CelestialNbodyGroupedGPU) {
		this->fallbackAlgorithm = NbodyAlgorithm::AllPairsGPU;
	} else {
		this->fallbackAlgorithm = UserInput::fallbackAlgorithm;
	}

	this->enableThresholdSwitching = UserInput::enableThresholdSwitching &&
			(simulationAlgorithm == NbodyAlgorithm::CelestialNbodyCPU ||
			 simulationAlgorithm == NbodyAlgorithm::CelestialNbodyGPU ||
			 simulationAlgorithm == NbodyAlgorithm::CelestialNbodyGroupedGPU ||
			 integrationMethod == IntegrationMethod::CelestialIntegrationExplicit ||
			 integrationMethod == IntegrationMethod::CelestialIntegrationSymplectic);

	if (useGPU) {

		if (renderSimulation) {
			position = newSystem->getPositions();
		} else {
			position = newSystem->getPositionsGPU();
		}

		velocity = newSystem->getVelocitiesGPU();
		acceleration = newSystem->getAccelerationsGPU();
		mass = newSystem->getMassesGPU();
	} else {

		position = newSystem->getPositions();
		velocity = newSystem->getVelocities();
		acceleration = newSystem->getAccelerations();
		mass = newSystem->getMasses();

	}

	renderingEnabled = false;
	currentTime = 0;
	systemVAO = 0;
	systemVBO = 0;
	positionCudaResource = 0;

}

CelestialSystem::CelestialSystem(float3 *position, float3 *acceleration,
		float3 *velocity, float* mass, int N, bool useGPU) {

	this->position = position;
	this->acceleration = acceleration;
	this->velocity = velocity;
	this->mass = mass;
	this->N = N;
	this->useGPU = useGPU;
	this->integrationMethod = UserInput::integrationMethod;
	this->simulationAlgorithm = UserInput::nbodyAlgorithm;
	this->enableThresholdSwitching = UserInput::enableThresholdSwitching;
	this->fallbackIntegrationMethod = IntegrationMethod::SymplecticEuler;
	this->thresholdReached = false;
	this->distanceThreshold = UserInput::distanceThreshold;
	this->currentTime = 0.0;
	this->renderingEnabled = false;
	this->renderSimulation = true;
	this->solarBodies = -1; //N/A
	this->solarSystems = -1; //N/A
	this->fallbackAlgorithm = UserInput::fallbackAlgorithm;

	//Variables used for rendering
	systemVBO = 0;
	systemVAO = 0;
	positionCudaResource = 0;

}

CelestialSystem::CelestialSystem(float3 *position, float3 *acceleration,
		float3 *velocity, float* mass, int solarSystems, int solarBodies,
		bool useGPU) {

	this->position = position;
	this->acceleration = acceleration;
	this->velocity = velocity;
	this->mass = mass;
	this->useGPU = useGPU;
	this->integrationMethod = UserInput::integrationMethod;
	this->simulationAlgorithm = UserInput::nbodyAlgorithm;
	this->fallbackIntegrationMethod = IntegrationMethod::SymplecticEuler;
	this->thresholdReached = false;
	this->enableThresholdSwitching = UserInput::enableThresholdSwitching;
	this->distanceThreshold = UserInput::distanceThreshold;
	this->currentTime = 0.0;
	this->renderingEnabled = false;
	this->renderSimulation = true;

	this->solarBodies = solarBodies;
	this->solarSystems = solarSystems;
	this->N = 1 + solarBodies * solarSystems;

	if (useGPU) {
		this->fallbackAlgorithm = NbodyAlgorithm::AllPairsGPU;
	} else {
		this->fallbackAlgorithm = NbodyAlgorithm::AllPairsCPU;
	}


	//Variables used for rendering
	systemVBO = 0;
	systemVAO = 0;
	positionCudaResource = 0;

}

CelestialSystem::~CelestialSystem() {

	if (useGPU) {

		if (!renderSimulation) {
			cudaFree(position);
		}
		cudaFree(acceleration);
		cudaFree(mass);
		cudaFree(velocity);
	} else {
		free(position);
		free(acceleration);
		free(mass);
		free(velocity);
	}

	renderingCleanup();
}

/* Getters and Setters */

float CelestialSystem::getKineticEnergy() {

	float *massCPU;
	float3 *velocityCPU;

	if (useGPU) {
		velocityCPU = (float3*) malloc(N * sizeof(float3));
		cudaMemcpy(velocityCPU, velocity, sizeof(float3) * N,
				cudaMemcpyDeviceToHost);

		massCPU = (float*) malloc(N * sizeof(float));
		cudaMemcpy(massCPU, mass, sizeof(float) * N,
				cudaMemcpyDeviceToHost);
	} else {
		massCPU = mass;
		velocityCPU = velocity;
	}

	float sumT = 0;
	for (int i = 0;i<N;i++) {

		float v_squared = velocityCPU[i].x * velocityCPU[i].x + velocityCPU[i].y * velocityCPU[i].y +
				velocityCPU[i].z * velocityCPU[i].z;

		sumT += 0.5 * massCPU[i] * v_squared;
	}

	if (useGPU) {
		free(massCPU);
		free(velocityCPU);
	}

	return sumT;

}

float CelestialSystem::getPotentialEnergy() {

	float *massCPU;
	float3 *velocityCPU;
	float3 *positionCPU;

	if (useGPU) {
		velocityCPU = (float3*) malloc(N * sizeof(float3));
		cudaMemcpy(velocityCPU, velocity, sizeof(float3) * N,
				cudaMemcpyDeviceToHost);
		positionCPU = (float3*) malloc(N * sizeof(float3));
		cudaMemcpy(positionCPU, position, sizeof(float3) * N,
				cudaMemcpyDeviceToHost);
		massCPU = (float*) malloc(N * sizeof(float));
		cudaMemcpy(massCPU, mass, sizeof(float) * N,
				cudaMemcpyDeviceToHost);
	} else {
		massCPU = mass;
		velocityCPU = velocity;
		positionCPU = position;
	}

	float U = 0;

	for (int i = 0;i<N; i++) {
		for (int j = 0;j<N; j++) {

			if (i == j) {
				continue;
			}

			float3 r = make_float3(positionCPU[i].x - positionCPU[j].x, positionCPU[i].y - positionCPU[j].y,
					positionCPU[i].z - positionCPU[j].z);
			float r_scalar = sqrt(r.x * r.x + r.y * r.y + r.z * r.z);

			U -= G * massCPU[i] * massCPU[j] / r_scalar;
		}
	}

	if (useGPU) {
		free(massCPU);
		free(velocityCPU);
		free(positionCPU);
	}

	return U;

}

void CelestialSystem::updateCurrentTime(float dt) {

	currentTime += dt;


}

int CelestialSystem::getN() {

	return this->N;

}

float CelestialSystem::getCurrentTime() {

	return currentTime;

}

int CelestialSystem::getSolarSystems() {

	return this->solarSystems;
}

int CelestialSystem::getSolarBodies() {

	return solarBodies;

}

bool CelestialSystem::isUsingGPU() {
	return this->useGPU;
}

float3 CelestialSystem::getBodyPosition(int body_i) {

	return this->position[body_i];

}

void CelestialSystem::setBodyPosition(int body_i, float x, float y, float z) {

	if (!useGPU) {
		this->position[body_i] = make_float3(x, y, z);
	}

}

float3 *CelestialSystem::getCPUAccelerations() { //Guaranteed to be in host memory

	float3 *acc;

	if (useGPU) {

		acc = (float3*) malloc(N * sizeof(float3));
		cudaMemcpy(acc, acceleration, sizeof(float3) * N,
				cudaMemcpyDeviceToHost);

	} else {
		acc = acceleration;
	}

	return acc;

}

float3 *CelestialSystem::getCPUPositions() {

	float3 *pos;

	if (useGPU) {

		pos = (float3*) malloc(N * sizeof(float3));
		cudaMemcpy(pos, position, sizeof(float3) * N,
				cudaMemcpyDeviceToHost);

	} else {
		pos = position;
	}

	return pos;
}

float3 *CelestialSystem::getPositions() {

	return position;

}

float3 *CelestialSystem::getAccelerations() {

	return acceleration;

}

float *CelestialSystem::getMasses() {

	return mass;

}

float3 *CelestialSystem::getVelocities() {

	return velocity;

}

NbodyAlgorithm CelestialSystem::getSimulationAlgorithm() {

	return simulationAlgorithm;

}

NbodyAlgorithm CelestialSystem::getFallbackAlgorithm() {

	return fallbackAlgorithm;

}

IntegrationMethod CelestialSystem::getIntegrationMethod() {

	return integrationMethod;

}

bool CelestialSystem::thresholdSwitchingIsEnabled() {

	return enableThresholdSwitching;

}

float CelestialSystem::getThresholdDistance() {

	return distanceThreshold;

}

bool CelestialSystem::thresholdIsReached() {

	return thresholdReached;

}

NbodyAlgorithm CelestialSystem::getCurrentlyEnabledAlgorithm() {

	NbodyAlgorithm currentAlgorithm;

	if (enableThresholdSwitching && thresholdReached) {
		currentAlgorithm = fallbackAlgorithm;
	} else {
		currentAlgorithm = simulationAlgorithm;
	}

	return currentAlgorithm;

}

IntegrationMethod CelestialSystem::getFallbackIntegrationMethod() {

	return fallbackIntegrationMethod;

}
IntegrationMethod CelestialSystem::getCurrentlyEnabledIntegrationMethod() {

	IntegrationMethod currentIntegrationMethod;

	if (enableThresholdSwitching && thresholdReached) {
		currentIntegrationMethod = fallbackIntegrationMethod;
	} else {
		currentIntegrationMethod = integrationMethod;
	}

	return currentIntegrationMethod;

}

void CelestialSystem::setThresholdReached(bool thresholdReached) {

	if (enableThresholdSwitching) {

		bool current = thresholdReached;
		bool previous = this->thresholdReached;

		if (previous != current) {
			if (current) {
				printf("[Simulation Output] Threshold %.2f Reached at time %.2f, continuing with fallback algorithm %s\n", distanceThreshold, currentTime,
						Utilities::NbodyEnumToText(fallbackAlgorithm).c_str());
			} else {
				printf("[Simulation Output] Below %.2f threshold, continuing with Celestial Algorithm\n", distanceThreshold);
			}
		}

		this->thresholdReached = current;

	}

}

void CelestialSystem::toggleIntegrationMethod() { //Toggle between celestial and symplectic integration

	bool enabled = false;

	if (integrationMethod == IntegrationMethod::SymplecticEuler) {
		integrationMethod = IntegrationMethod::CelestialIntegrationSymplectic;
		enabled = true;
	} else if (integrationMethod == IntegrationMethod::CelestialIntegrationSymplectic) {
		integrationMethod = IntegrationMethod::SymplecticEuler;
	}

	printf("[KEYBOARD INPUT] CelestialIntegration %s\n", enabled ? "Enabled" : "Disabled");

}

void CelestialSystem::printEnergyValues() {

	float T = getKineticEnergy();
	float U = getPotentialEnergy();

	printf("[SIMULATION OUTPUT] Kinetic Energy: %E\n", T);
	printf("[SIMULATION OUTPUT] Potential Energy: %E\n", U);
	printf("[SIMULATION OUTPUT] Total Energy: %E\n", T+U);

}



