/*
 * CelestialSystem.h
 *
 *  Created on: Feb 18, 2015
 *      Author: Joakim
 */

#ifndef CELESTIALSYSTEM_H_
#define CELESTIALSYSTEM_H_

#include <GL/glew.h>
#include <GL/gl.h>
#include <SDL2/SDL_opengl.h>
#include <cuda_gl_interop.h>
#include "../math/integration/IntegrationMethod.h"
#include "../nbody/nbodyAlgorithm.h"

class CelestialSystem {
private:

	//Body variables
	float3 *position;
	float3 *acceleration;
	float3 *velocity;
	float *mass;

	//Bodies Information
	int N; //Total
	int solarSystems; //Number of solar systems
	int solarBodies; //Number of bodies in the solar system (including the sun)

	//Simulation flags and variables
	bool useGPU;
	IntegrationMethod integrationMethod;
	NbodyAlgorithm simulationAlgorithm;
	NbodyAlgorithm fallbackAlgorithm;
	IntegrationMethod fallbackIntegrationMethod;
	bool enableThresholdSwitching;
	float distanceThreshold;
	bool thresholdReached;

	bool renderingEnabled; //Signals that the system is ready for rendering
	bool renderSimulation; //Makes necessary preparations in order to associate it with a renderer
	float currentTime;

	//Rendering Variables
	GLuint systemVBO;
	GLuint systemVAO;
	cudaGraphicsResource_t positionCudaResource;

	void performIntegration(float dt);
	void renderingCleanup();
	void mapResourcesCuda();
	void unmapResourcesCuda();
	void updateBufferCPU();
	void updateCurrentTime(float dt);

public:
	CelestialSystem(float3 *position, float3 *acceleration, float3 *velocity, float* mass, int N, bool useGPU);
	CelestialSystem(float3 *position, float3 *acceleration, float3 *velocity, float* mass, int solarSystems, int solarBodies, bool useGPU);
	CelestialSystem(int solarSystems, int solarBodies, bool useGPU, bool renderSimulation);
	CelestialSystem(int solarSystems, int solarBodies, bool useGPU);
	~CelestialSystem();
	void moveBodies(float dt);
	void drawBodies();
	void setupForRendering();
	void setBodyPosition(int body_i, float x, float y, float z);
	//Getters
	float getCurrentTime();
	float3 *getPositions();
	float3 *getAccelerations();
	float *getMasses();
	float3 *getVelocities();
	int getN();
	int getSolarSystems();
	int getSolarBodies();
	bool isUsingGPU();
	float3 getBodyPosition(int body_i);
	float3 *getCPUAccelerations();
	float3 *getCPUPositions();
	NbodyAlgorithm getSimulationAlgorithm();
	NbodyAlgorithm getFallbackAlgorithm();
	NbodyAlgorithm getCurrentlyEnabledAlgorithm();
	IntegrationMethod getIntegrationMethod();
	IntegrationMethod getFallbackIntegrationMethod();
	IntegrationMethod getCurrentlyEnabledIntegrationMethod();
	bool thresholdSwitchingIsEnabled();
	bool thresholdIsReached();
	float getThresholdDistance();
	void setThresholdReached(bool thresholdReached);
	void toggleIntegrationMethod();
	void printEnergyValues();
	float getKineticEnergy();
	float getPotentialEnergy();

};

#endif /* CELESTIALSYSTEM_H_ */
