#include "CelestialTestcase.h"
#include "../system/CelestialSystem.h"
#include "../math/comparison/ValueComparison.h"
#include "../input/UserInput.h"
#include "../utility/RandomNumberGenerator.h"
#include "../utility/CudaTimer.cuh"
#include "../utility/Utilities.h"
#include <stdio.h>
#include <string>

void CelestialTestcase::nbodyAccelerationComparisonTest(NbodyAlgorithm nbodyAlgorithm_a, NbodyAlgorithm nbodyAlgorithm_b,
		IntegrationMethod integrationMethod_a, IntegrationMethod integrationMethod_b, int solarSystems, int solarBodies,
		float dt, int steps, bool verbose, bool showInfoByStep, bool showGraph) {

	//Test algorithm a
	bool useGPU_A = UserInput::setupSimulation(nbodyAlgorithm_a, integrationMethod_a);
	RandomNumberGenerator::setSrand(0);
	CelestialSystem *systemA = new CelestialSystem(solarSystems,solarBodies, useGPU_A, false);

	//Test algorithm b
	bool useGPU_B = UserInput::setupSimulation(nbodyAlgorithm_b, integrationMethod_b);
	RandomNumberGenerator::setSrand(0);
	CelestialSystem *systemB = new CelestialSystem(solarSystems,solarBodies, useGPU_B, false);

	int N = 1 + solarSystems * solarBodies;

	printf("------------------------------------------------------------------------------------------\n");
	printf("Comparison Test Case: SolarSystems %d, SolarBodies %d -- dt = %.2f\n",
			solarSystems, solarBodies, dt);
	printf("------------------------------------------------------------------------------------------\n");


	std::string nameA = Utilities::NbodyEnumToText(nbodyAlgorithm_a);
	std::string nameB = Utilities::NbodyEnumToText(nbodyAlgorithm_b);

	printf("------------------------------------------------------------------------------------------\n");
	printf("System A: %s      System B: %s    Time period: %.f    dt: %.f\n", nameA.c_str(), nameB.c_str(), (steps+1) * dt, dt);
	printf("------------------------------------------------------------------------------------------\n");

	float *xArray, *yArray;
	int dataLength;
	int dataOffset;

	if (showGraph) {
		dataOffset = 0;
		dataLength  = steps;
		xArray = (float*) malloc(sizeof(float) * dataLength );
		yArray = (float*) malloc(sizeof(float) * dataLength);
	}

	for (int i=0;i<steps;i++) {


		float3 *pos_A = systemA->getCPUPositions();
		float3 *pos_B = systemB->getCPUPositions();
		float posDifflen = ValueComparison::compareArrayValues3f(pos_A, pos_B, N, verbose);
		//float posDifflen = ValueComparison::medianValue(pos_A, pos_B, N);

		float3 *acc_A = systemA->getCPUAccelerations();
		float3 *acc_B = systemB->getCPUAccelerations();

		if (showInfoByStep) {
			float accDifflen = ValueComparison::compareArrayValues3f(acc_A, acc_B, N, verbose);
			float accBigDiff = ValueComparison::biggestDifference3f(acc_A, acc_B, N);
			float posBigDiff = ValueComparison::biggestDifference3f(pos_A, pos_B, N);
			//float T_A = systemA->getKineticEnergy();
			//float T_B = systemB->getKineticEnergy();
			//float U_A = systemA->getPotentialEnergy();
			//float U_B = systemB->getPotentialEnergy();
			//float E_A = T_A + U_A;
			//float E_B = T_B + U_B;

			printf("------------------------------------------------------------------------------------------\n");
			printf("Comparison at time %.f days\n", dt * i);
			printf("------------------------------------------------------------------------------------------\n");
			printf("Acceleration Absolute Difference:  %E  Biggest Difference: %E\n", accDifflen, accBigDiff);
			printf("Position Absolute Difference:      %E  Biggest Difference: %E\n", posDifflen, posBigDiff);
			printf("------------------------------------------------------------------------------------------\n");

			/*
			printf("Energy Values\n");
			printf("------------------------------------------------------------------------------------------\n");
			printf("Kinetic Energy A: %E\n", T_A);
			printf("Kinetic Energy B: %E\n", T_B);
			printf("Potential Energy A: %E\n", U_A);
			printf("Potential Energy B: %E\n", U_B);
			printf("Total Energy A: %E\n", E_A);
			printf("Total Energy B: %E\n", E_B);
			printf("------------------------------------------------------------------------------------------\n");
			printf("\n");
			*/
		}

		//Save Results in array
		if (posDifflen > 0) {
			xArray[i] = (i * dt) / 365.0;
			yArray[i] = log10(posDifflen);
		} else {
			dataOffset++;
			dataLength--;
		}

		systemA->moveBodies(dt);
		systemB->moveBodies(dt);

		if (useGPU_A) {
			free(acc_A);
			free(pos_A);
		}

		if (useGPU_B) {
			free(acc_B);
			free(pos_B);
		}

	}

	if (dataLength > 0) {

		if (showGraph) {

			std::string xlabel = "xlabel = 'X axis: time in years (dt = " + std::to_string(dt) + " days)'";
			std::string ylabel = "ylabel = 'Y axis: The total absolute distance difference in AU (Logarithmic)'";
			std::string title = "title = '" + std::string(Utilities::NbodyEnumToText(nbodyAlgorithm_a)) + " and " +
					std::string(Utilities::NbodyEnumToText(nbodyAlgorithm_b)) + " comparison(" + std::to_string(solarSystems) +
					"," + std::to_string(solarBodies) + ")'";
			Utilities::runPythonPlotScript(xArray, yArray, dataLength, dataOffset, xlabel.c_str(), ylabel.c_str(), title.c_str());
			free(xArray);
			free(yArray);
		}
	} else {
		printf("[SIMULATION OUTPUT] No measurable difference between the algorithms\n");
	}

	delete systemA;
	delete systemB;

}

void CelestialTestcase::nbodyPerformanceTest(NbodyAlgorithm nbodyAlgorithm, IntegrationMethod integrationMethod, int solarSystems, int solarBodies, float dt) {

	if (UserInput::logSimulation) {
		printf("[LOG] Starting Performance test\n");
	}

	float time;

	bool useGPU = UserInput::setupSimulation(nbodyAlgorithm, integrationMethod);
	RandomNumberGenerator::setSrand(0);
	CelestialSystem *system = new CelestialSystem(solarSystems,solarBodies, useGPU, false);
	CudaTimer *timer = new CudaTimer();
	timer->cudaTimerStart();
	system->moveBodies(dt);
	time = timer->cudaTimerStop();

	delete system;

	std::string name = Utilities::NbodyEnumToText(nbodyAlgorithm);
	std::string integration = Utilities::IntegrationEnumToText(integrationMethod);

	printf("Performance Test Case: SolarSystems %d, SolarBodies %d -- dt = %f\n",
			solarSystems, solarBodies, dt);
	printf("----------------------------\n");
	printf("%s with %s time: %f ms\n", name.c_str(), integration.c_str(), time);
	printf("\n");

}

void CelestialTestcase::nbodyPerformanceComparisonTest(NbodyAlgorithm nbodyAlgorithm_a, NbodyAlgorithm nbodyAlgorithm_b, IntegrationMethod integrationMethod_a, IntegrationMethod integrationMethod_b,
		int solarSystems, int solarBodies, float dt) {

	nbodyPerformanceTest(nbodyAlgorithm_a, integrationMethod_a, solarSystems, solarBodies, dt);
	nbodyPerformanceTest(nbodyAlgorithm_b, integrationMethod_b, solarSystems, solarBodies, dt);

}
